// cohorte.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-09 15:13:21 tac>
//
// 年齢等の時期項目別コホート分析を行う。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -W cohorte cohorte.C
//               -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "cohorte  Version 7.41e       Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:cohorte  [<OPTION> ...]  [<DATA FILE> ...]",
  "  The cohort analysis with the first item as period and the second and",
  "  subsequent items as analysis data.",
  "<OPTION>",
  "  -dt=<INTERVAL>:specify time interval(default:-dt=1)",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -ngp={msd|mse}[,...]:select the data for Ngraph",
  "    msd:mean +/- SD  mse:mean +/- SE  default:all",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=cohorte_ngp.dat",
  "  -res=[-][{def|all|statc|statt}][,...]:select output results",
  "    -:wiht -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    statc:cohort basic statistics  statt:overall basic statistics",
  "  -t=<TIME1>,<TIME2>[,<TIME3>,...]:specify the relative time of items",
  "    default:-t=0,1,...,(number of items-1)"
};

const string Title = "=== Cohort Analysys ==="; // 表題
const string NGPNAMEdef = "cohorte_ngp.dat"; // 省略時Ngraphファイル名

using vecstatv = vector<statv<double>>; // 基礎統計量ベクトルの短名定義

enum ngptype                    // Ngraph用出力データの種類
{
  ngpnon,                       // 出力なし
  ngpmsd,                       // 平均値±SD
  ngpmse,                       // 平均値±SE
  ngpetc                        // その他
};

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vdt;                   // 時期間隔
  string vngpname;              // Ngraphファイル基本名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  lstr tlst;                    // 項目相対時間リスト
  vecdbl vtvec;                 // 項目相対時間ベクトル
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double dt() const             // 時期間隔を返す
  { return vdt; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 分析結果出力フラッグを返す
  double t(size_t idx) const    // 項目時間を返す
  { return (idx < vtvec.size()) ? vtvec[idx] : BLANKf; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vdt(1.0), vngpname(), vngplst(), vngpmap(), vreslst(), vresmap(), tlst(), vtvec()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngpitr = vngpmap.find(code);
  if (ngpitr == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "dt")           // 時期間隔の指定
  {
    vdt = strtodf(opt[1]);
    if (vdt <= 0.0 || vdt == BLANKf) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "t")       // 項目時間の指定
  {
    if (strsplit(tlst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのチェックとセット
{
  if (ni < 2) return disperr(vpname, "At least 2 items are required !");
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  vtvec.resize(ni);
  for (size_t j = 0; j < ni; ++j) vtvec[j] = (double)j * vdt;
  if (settvec(vtvec, tlst, ni) < 0) return disperr(vpname, "Wrong relative time of items !");
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["msd"] = vngpmap["mse"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["msd"] = vngpmap["mse"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngpitr = vngpmap.find(ngp);
      if (ngpitr == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["statc"] = vresmap["statt"] = true;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["statc"] = vresmap["statt"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def" || res == "all")
      {
        vresmap["def"] = vresmap["statc"] = vresmap["statt"] = true;
      }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

using mapdv = map<double, vecstatv>; // コホート別統計量マップの別名定義
using mapdv_itr = mapdv::iterator; // 反復子の別名定義
using mapdv_citr = mapdv::const_iterator; // const反復子の別名定義

class cohortclass               // cohortメインクラス
{
private:
  option opt;                   // オプション
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  void vecstatv_sigma(vecstatv& statvec, datac& cdata); // データを累積する
  void put_resid(ostream& fout); // 解析結果のタイトルとデータ識別情報を出力する
  void put_chlist(ostream& fout, mapdv& vecstatvmap);
  // コホート表を出力する
  void sum_yt(double tmin, vecstatv& ytvec, mapdv& vecstatvmap) const;
  // コホートを加算する
  void plot_ngp(mapdv& vecstatvmap); // Ngraph用データを出力する
  void plot_ngpid(ofstream& fout);
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_msde(ofstream& fout, ngptype ngp, mapdv& vecstatvmap) const;
  // Ngraph用に平均値±SD/SEを出力する
public:
  cohortclass(const char* c_pname = nullptr) : opt(c_pname), fname(), id(), datalst()
  // コンストラクタ
  {}
  ~cohortclass()                // デストラクタ
  {}
  const string& pname() const   // プログラム名を返す
  { return opt.pname(); }
  void init()      // 初期設定を行う(コンストラクタ中ではできない処理)
  { opt.setdef(); }
  void setopt(const string& optstr, char devc = OPTdev) // オプションを設定する
  { opt.set(optstr, devc); }
  int get(const char* c_fname); // データを読み込み累積する
  void put(int nc);             // 集計結果を出力する
};

void cohortclass::vecstatv_sigma(vecstatv& statvec, datac& cdata)
// データを累積する
{
  int idxend = min(statvec.size() + 1, cdata.n());
  for (int i = 1; i < idxend; ++i)
  {
    double fdata = cdata.fdata(i);
    if (fdata != BLANKf) statvec[i - 1].sigma(fdata);
  }
}

void cohortclass::put_resid(ostream& fout)
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "COHORT ITEM    :" << id.item(0) << "\n";
  fout << "PERIOD ITEM   1:" << id.item(1) << "\n";
  fout << "  :\n";
  fout << "PERIOD ITEM" << setw(4) << id.ni() - 1;
  fout << ":" << id.item(id.ni() - 1) << "\n";
}

void cohortclass::put_chlist(ostream& fout, mapdv& vecstatvmap)
// コホート表を出力する
{
  constexpr int LENline = 79;   // 区切り線の長さ
  double tmin = 0.0, tmax = 0;
  int k = 0;
  for (mapdv_itr chitr = vecstatvmap.begin(); chitr != vecstatvmap.end(); ++chitr)
  {
    int nd = 0;
    for (size_t j = 0; j < chitr->second.size(); ++j)
    {
      if ((chitr->second[j]).n() > 0) ++nd;
    }
    if (nd > 0)
    {
      ++k;
      if (k == 1) tmin = chitr->first;
      if (opt.res("statc"))
      {
        fout << "\n<COHORT " << k << ":Start Time t=" << chitr->first << ">\n";
        putline(fout, LENline);
      }
      double t = 0.0;
      for (size_t j = 0; j < chitr->second.size(); ++j)
      {
        statv<double>* yj = &(chitr->second[j]);
        if (yj->n() > 0)
        {
          t = chitr->first + opt.t(j);
          if (opt.res("statc"))
          {
            fout << "t=" << setw(4) << t << ":";
            putmsd(fout, yj->n(), yj->mean(), yj->vr());
          }
        }
      }
      if (opt.res("statc")) putline(fout, LENline);
      if (t > tmax) tmax = t;
    }
  }
  if (opt.res("statt") == false) return;
  int p = int(tmax - tmin + 0.99) + 1;
  vecstatv ytvec(p);
  sum_yt(tmin, ytvec, vecstatvmap);
  fout << "\n<OVERALL:Start Time t=" << tmin << ">\n";
  putline(fout, LENline);
  for (int i = 0; i < p; ++i)
  {
    if (ytvec[i].n() > 0)
    {
      fout << "t=" << setw(4) << tmin + (double)i * opt.dt() << ":";
      putmsd(fout, ytvec[i].n(), ytvec[i].mean(), ytvec[i].vr());
    }
  }
  putline(fout, LENline);
}

void cohortclass::sum_yt(double tmin, vecstatv& ytvec, mapdv& vecstatvmap) const
// コホートを加算する
{
  for (mapdv_itr chitr = vecstatvmap.begin(); chitr != vecstatvmap.end(); ++chitr)
  {
    for (size_t j = 0; j < chitr->second.size(); ++j)
    {
      if ((chitr->second[j]).n() > 0)
      {
        int idx = (int)(chitr->first + opt.t(j) - tmin);
        ytvec[idx] += chitr->second[j];
      }
    }
  }
}

void cohortclass::plot_ngp(mapdv& vecstatvmap) // Ngraph用データを出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("mse"))           // 平均値±SE
  {
    fout << "-s1 -dline -w80\n";
    plot_ngpid(fout);
    plot_msde(fout, ngpmse, vecstatvmap);
    delml = true;
  }
  if (opt.ngp("msd"))           // 平均値±SD
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dline -w80\n";
    plot_ngpid(fout);
    plot_msde(fout, ngpmsd, vecstatvmap);
  }
}

void cohortclass::plot_ngpid(ofstream& fout)
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%When Draw SD or SE:-derrorbar_y -w40 -o500\n";
  fout << "%";  puttitle(fout, Title);  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%COHORT ITEM    :" << id.item(0) << "\n";
  fout << "%PERIOD ITEM   1:" << id.item(1) << "\n";
  fout << "%  :\n";
  fout << "%PERIOD ITEM" << setw(4) << id.ni() - 1;
  fout << ":" << id.item(id.ni() - 1) << "\n";
}

void cohortclass::plot_msde(ofstream& fout, ngptype ngp, mapdv& vecstatvmap) const
// Ngraph用平均値±SD/SE出力
{
  fout << "%<MEAN +/- ";
  if (ngp == ngpmsd) fout << "SD";
  else fout << "SE";
  fout << ">\n";
  double tmin = 0.0;
  int k = 0;
  for (mapdv_itr chitr = vecstatvmap.begin(); chitr != vecstatvmap.end(); ++chitr)
  {
    if (k > 0) fout << "= =\n";
    fout << "%COHORT " << k + 1 << "\n";
    int nd = 0;
    for (size_t j = 0; j < chitr->second.size(); ++j)
    {
      if ((chitr->second[j]).n() > 0) ++nd;
    }
    if (nd > 0)
    {
      ++k;
      if (k == 1) tmin = chitr->first;
      double t = 0.0;
      for (size_t j = 0; j < chitr->second.size(); ++j)
      {
        statv<double>* yj = &(chitr->second[j]);
        if (yj->n() > 0)
        {
          t = chitr->first + opt.t(j);
          double ebar = (ngp == ngpmsd) ? yj->sd() : yj->se();
          fout << t << " " << yj->mean() << " " << -ebar << " " << ebar << endl;
        }
      }
    }
  }
}

int cohortclass::get(const char* c_fname) // データを読み込み累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void cohortclass::put(int nc)   // 集計結果を出力する
{
  if (nc < 1)                   // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout);
      cout << "\n" << endl;
    }
    return;
  }
  mapdv vecstatvmap;            // コホート別統計量ベクトルマップ
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    double t = ditr->fdata(0);
    if (t != BLANKf)
    {
      mapdv_itr chitr = vecstatvmap.find(t);
      if (chitr == vecstatvmap.end()) // 新規コホート
      {
        vecstatv statvec(ditr->n() - 1);
        for (size_t i = 1; i < ditr->n(); ++i)
        {
          double fdata = ditr->fdata(i);
          if (fdata != BLANKf) statvec[i - 1].sigma(fdata);
        }
        vecstatvmap.insert(map<double, vecstatv>::value_type(t, statvec));
      }
      else                      // 既存のコホート
      { vecstatv_sigma(chitr->second, *ditr); }
    }
  }
  if (opt.res("res"))           // 計算結果出力
  {
    put_resid(cout);
    put_chlist(cout, vecstatvmap);
    cout << "\n" << endl;
  }
  if (opt.ngpname() != "") plot_ngp(vecstatvmap);
}

int main(int argc, char* argv[]) // メインルーチン
{
  cohortclass cohort(argv[0]);  // メインクラス変数
  try
  {
    cohort.init();              // cohortの初期設定
    bool fok = true;            // ファイル処理フラッグ
    int nc = 0;                 // 読み込んだケース数
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) nc += cohort.get(nullptr); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { cohort.setopt(argv[i] + 1);  fok = true; }
      else
      { nc += cohort.get(argv[i]);  fok = false; }
    }
    cohort.put(nc);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(cohort.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(cohort.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
