// cosinore.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-11 10:45:13 tac>
//
// コサイナー法によるケースごとの周期回帰分析を行う。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -w cosinore cosinore.C
//               -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "cosinore  Version 7.41e      Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:cosinor  [<OPTION> ...]  [<DATA FILE> ...]",
  "  The periodic regression analysis by cosinor method.",
  "<OPTION>",
  "  -dnsname[=<FILE>]:specify the file to outout DANS data",
  "    default:no output  -dnsname:-dnsname=cosinore_dns.d",
  "  -dt=<INTERVAL>:specify time interval(default:-dt=T/number of periods)",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -k=<NUMBER>[,<NUMBER>,...]:specify the periodic component number to be",
  "    included in the periodic regression equation",
  "  -ngp=<MAX>:specify coordinate axis maximums for Ngraph(0<MAX)",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=cosinore_ngp.dat",
  "  -res=[-][{def|all|cc[<C-COEF>]|cos}][,...]:select output results",
  "    -:wiht -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    cc[<C-COEF>]:confidence interval(0=<C=COEF<1)  cos:cosinor method",
  "  -T=<PERIOD>:specify the basic period(default:-T=24)",
  "  -t=<TIME1>,<TIME2>[,<TIME3>,...]:specify the relative time of items",
  "    default:-t=0,T/(number of periods),...,",
  "    (number of periods-1)xT/(number of periods)"
};

const string Title = "=== Cosinor Method ==="; // 表題
const string DNSNAMEdef = "cosinore_dns.d"; // 省略時DANSファイル名
const string NGPNAMEdef = "cosinore_ngp.dat"; // 省略時Ngraphファイル名
constexpr double T0def = 24.0;  // 省略時基本周期
constexpr int LENline = 76;     // 区切り線の長さ

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  string vdnsname;              // DANSファイル名
  double vdt;                   // 時期間隔
  double vamax;                 // 座標軸最大値
  string vngpname;              // Ngraphファイル名
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  lstr klst;                    // 周期成分番号リスト
  vecint vpcvec;                // 周期成分番号配列
  int vpcmax;                   // 周期成分番号最大値
  double vt0;                   // 基本周期
  lstr tlst;                    // 項目時間リスト
  vecdbl vtvec;                 // 項目時間ベクトル
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  const string& dnsname() const // DANSデータファイル名を返す
  { return vdnsname; }
  double dt() const             // 時期間隔を返す
  { return vdt; }
  int m() const                 // 周期成分数を返す
  { return vpcvec.size(); }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  const vecint& pcvec() const   // 周期成分番号ベクトルを返す
  { return vpcvec; }
  int pcn(size_t k) const       // 周期成分番号を返す
  { return (k < vpcvec.size()) ? vpcvec[k] : 0; }
  int pcmax() const             // 周期成分番号最大値を返す
  { return vpcmax; }
  double amax() const           // 座標軸最大値を返す
  { return vamax; }
  bool res(const string& code); // 分析結果出力フラッグを返す
  double t0() const             // 基本周期を返す
  { return vt0; }
  double t(size_t idx) const    // 項目時間を返す
  { return (idx < vtvec.size()) ? vtvec[idx] : vt0; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vdnsname(), vdt(0.0), vamax(BLANKf), klst(), vngpname(), vpcvec(), vpcmax(-1), vreslst(), vresmap(), vt0(T0def), tlst(), vtvec()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "dnsname")      // DANSファイル名指定
  {
    if (opt[1].empty()) vdnsname = DNSNAMEdef;
    else vdnsname = opt[1];
    if (vdnsname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "dt")      // 時期間隔の指定
  {
    vdt = strtodf(opt[1]);
    if (vdt <= 0.0 || vdt == BLANKf) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "k")       // 周期成分番号の指定
  {
    if (strsplit(klst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngp")     // Ngraph用座標軸最大値の指定
  {
    vamax = opt[1].empty() ? BLANKf : strtodf(opt[1]);
    if (vamax <= 0.0) err = badopt(vpname, optstr, USAGE, 0.0, DBL_MAX);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "T")       // 基本周期の指定
  {
    vt0 = strtodf(opt[1]);
    if (vt0 == BLANKf || vt0 <= 0.0) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "t")       // 項目時間の指定
  {
    if (strsplit(tlst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのチェックとセット
{
  if (ni < 3) throw runtime_error("At least 3 items are required !");
  if (set_res() < 0) return -3; // -resオプションのセット
  int p = ni;
  int q = ni / 2;                // 周期成分最大数
  vtvec.resize(ni);
  double dt = (vdt > 0.0) ? vdt : vt0 / (double)ni;
  for (size_t j = 0; j < ni; ++j) vtvec[j] = (double)j * dt;
  if (settvec(vtvec, tlst, ni) < 0) return disperr(vpname, "Wrong relative time of items !");
  if (klst.empty())             // 周期成分指定がない時
  {
    vpcvec.resize(1);
    vpcvec[0] = 1;              // 基本周期だけ選択
  }
  else if ((vpcmax = setpcvec(vpcvec, klst, q)) < 0)
  { return disperr(vpname, "Wrong periodic component !"); }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["cos"] = true;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["cc"] = vresmap["cos"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def" || res == "all")
      {
        vresmap["def"] = vresmap["cc"] = vresmap["cos"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class cosdata                   // コサイナー用データクラス
{
private:
  int vcn;                      // ケース番号
  int vn;                       // データ数
  int vnc;                      // ケース数
  mat<double> vsmat;            // 計算用行列
  double vsr;                   // 残差
  double vsyy;                  // 平方和
  vecdbl sigvec;                // Σ1,x,yベクトル(グループ累積用)
public:
  cosdata() : vcn(0), vn(0), vnc(0), vsmat(), vsr(0.0), vsyy(0.0), sigvec()
  // デフォールトコンストラクタ
  {}
  cosdata(int cn0, int d) : vcn(cn0), vn(0), vnc(0), vsmat(d), vsr(0.0), vsyy(0.0), sigvec(d)
  // コンストラクタ
  {}
  cosdata(const cosdata& cosdata0) : vcn(cosdata0.vcn), vn(cosdata0.vn), vnc(cosdata0.vnc), vsmat(cosdata0.vsmat), vsr(cosdata0.vsr), vsyy(cosdata0.vsyy), sigvec(cosdata0.sigvec)
  // コピーコンストラクタ
  {}
  virtual ~cosdata()            // デストラクタ
  {}
  const cosdata& operator=(const cosdata& cosdata0); // コピー代入演算子
  const cosdata& operator+=(const cosdata& cosdata0); // 加算代入演算子
  const cosdata& reset(int cn0, int d); // データを再設定する
  int cn() const                        // ケース番号を返す
  { return vcn; }
  int n() const                 // データ数を返す
  { return vn; }
  int n(int n0)                 // データ数を設定する
  { return (vn = n0); }
  int nc() const                // ケース数を返す
  { return vnc; }
  int nc(int nc0)               // ケース数を設定する
  { return (vnc = nc0); }
  double smat(int i, int j) const // 計算用行列要素を返す
  { return vsmat.at(i, j); }
  double smat_add(int i, int j, double f) // 計算用行列要素に加算する
  { return vsmat.add(i, j, f); }
  double sr() const             // 残差を返す
  { return vsr; }
  double sr(double sr0)         // 残差を設定する
  { return (vsr = sr0); }
  double syy() const            // 平方和を返す
  { return vsyy; }
  double syy(double syy0)       // 平方和を設定する
  { return (vsyy = syy0); }
  void cal_pfun();              // 周期関数を計算する
  void cal_sxy();               // 共分散行列を計算する
  void cal_gpfun();             // グループ周期関数を計算する
  void cal_gsxy();              // グループ共分散行列を計算する
};

const cosdata& cosdata::operator=(const cosdata& cosdata0) // コピー代入演算子
{
  if (&cosdata0 == this) return *this; // 自分自身への代入を防ぐ
  vcn = cosdata0.vcn;  vn = cosdata0.vn;  vnc = cosdata0.vnc;
  vsmat = cosdata0.vsmat;  vsr = cosdata0.vsr;  vsyy = cosdata0.vsyy;
  sigvec = cosdata0.sigvec;
  return *this;
}

const cosdata& cosdata::operator+=(const cosdata& cosdata0) // 加算代入演算子
{
  vn += cosdata0.vn;  vnc += cosdata0.vnc;  vsmat += cosdata0.vsmat;
  for (size_t i = 0; i < sigvec.size(); ++i) sigvec[i] += cosdata0.sigvec[i];
  return *this;
}

const cosdata& cosdata::reset(int cn0, int d) // データ再設定
{
  vcn = cn0;  vn = 0;  vnc = 0;  vsmat.resize(d, d);
  vsr = 0.0;  vsyy = 0.0;  sigvec.resize(d);
  vsmat = 0.0;  sigvec.assign(d, 0.0);
  return *this;
}

void cosdata::cal_pfun()        // 周期関数を計算する
{
  int d = vsmat.r();
  for (int j = 1; j < d - 1; ++j) vsmat.sweep(j);
  const double Veps = DBL_EPSILON * fmax(5.0, vsmat.maxabse());
  if (vsmat(d - 1, d - 1) < Veps) vsmat.assign(d - 1, d - 1, 0.0);
  vsr = vsmat(d - 1, d - 1);
}

void cosdata::cal_sxy()         // 共分散行列を計算する
{
  int d = vsmat.r();
  for (int i = 0; i < d - 1; ++i)
  {
    for (int j = i + 1; j < d; ++j) vsmat.assign(j, i, vsmat(i, j));
  }
  vn = (int)(vsmat(0, 0));
  for (int j = 0; j < d; ++j) sigvec[j] = vsmat(0, j);
  vsmat.sweep(0);
  vsyy = vsmat(d - 1, d - 1);
}

void cosdata::cal_gpfun()       // グループ周期関数を計算する
{
  int d = vsmat.r();
  for (int j = 1; j < d - 1; ++j) vsmat.sweep(j);
  const double Veps = DBL_EPSILON * fmax(5.0, vsmat.maxabse());
  if (vsmat(d - 1, d - 1) < Veps) vsmat.assign(d - 1, d - 1, 0.0);
}

void cosdata::cal_gsxy()        // グループ共分散行列を計算する
{
  int d = vsmat.r();
  vsmat.assign(0, 0, 1.0 / sigvec[0]);
  for (int i = 1; i < d; ++i)
  {
    vsmat.assign(0, i, sigvec[i] / sigvec[0]);
    vsmat.assign(i, 0, vsmat(0, i));
  }
  vsyy = vsmat(d - 1, d - 1);
}

class cosinorclass              // cosinorメインクラス
{
private:
  option opt;                   // オプション
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  void put_resid(ostream& fout); // 解析結果のタイトルとデータ識別情報を出力する
  void put_title(ostream& fout); // コサイナー法のタイトルを出力する
  void put_dnsid(ofstream& fout);
  // DANSデータのタイトルとデータ識別情報を出力する
  void cal_cos(cosdata& cosdg, cosdata& cosdc, datac& cdata);
  // コサイナー法の計算
  void put_dns(ofstream& fout, const cosdata& cosdc) const;
  // 振幅と位相をデータファイル形式で出力する
  void put_cosinor(ostream& fout, const cosdata& cosd, bool groupc);
  // コサイナー法結果の出力
  void plot_ngpcos(const cosdata& cosdg);
  // Ngraph用にコサイナーデータを出力する
  void plot_ngpid(ofstream& fout);
  // Ngraph用データのタイトルとデータ識別情報を出力する
public:
  cosinorclass(const char* c_pname) : opt(c_pname), fname(), id(), datalst()
  // コンストラクタ
  {}
  ~cosinorclass()               // デストラクタ
  {}
  const string& pname() const   // プログラム名を返す
  { return opt.pname(); }
  void init()      // 初期設定を行う(コンストラクタ中ではできない処理)
  { opt.setdef(); }
  void setopt(const string& optstr, char devc = OPTdev) // オプションを設定する
  { opt.set(optstr, devc); }
  int get(const char* c_fname); // データを読み込み累積する
  void put(int nc);             // 集計結果を出力する
};

void cosinorclass::put_resid(ostream& fout)
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "PERIOD ITEM  1:" << id.item(0) << "\n";
  fout << "  :\n";
  fout << "PERIOD ITEM" << setw(3) << id.ni();
  fout << ":" << id.item(id.ni() - 1) << "\n";
}

void cosinorclass::put_title(ostream& fout)
// 解析結果のタイトルを出力する
{
  fout << "\n<PERIODIC REGRESSION EQUATION>\n";
  fout << "y = A0 + sum{Ak x cos(wk x t - hk)}\n";
  fout << "k=1,…,p:Periodic Component  A0:MESOR  Ak:Amplitude\n";
  fout << "wk=360deg x k/T:Angular Frequency  hk:Acrophase  T=" << opt.t0() << ":basic period\n";
  fout << "T/k:Period  f=1/T:Basic Frequency  fk=k/T:Frequency\n";
  fout << "t=";
  for (size_t j = 0; j < id.ni(); ++j)
  {
    if (j > 1 && j % 12 == 1) fout << "\n  ";
    fout << setw(6) << opt.t(j);
  }
  fout << "\nCase No.    A0        Ak     hk/wk R-squared   F-value DF1 DF2   p-value\n";
  putline(fout, LENline);
}

void cosinorclass::put_dnsid(ofstream& fout)
// DANSデータのタイトルとデータ識別情報を出力する
{
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
  idd tid(id.title(), opt.m() * 2 + 1); // 項目名出力用
  tid.assign_item(0, "MESOR");
  ostringstream ostr;
  for (int k = 0; k < opt.m(); ++k)
  {
    double t = opt.t0() / opt.pcn(k);
    ostr.str("");
    ostr << "Amplitude " << t << " Time Period";
    int idx = 2 * (k + 1) - 1;
    tid.assign_item(idx, ostr.str());
    ostr.str("");
    ostr << "Acrophase " << t << " Time Period";
    tid.assign_item(idx + 1, ostr.str());
  }
  tid.put_item(fout, formtag);
}

void cosinorclass::cal_cos(cosdata& cosdg, cosdata& cosdc, datac& cdata)
// コサイナー法の計算
{
  int p = id.ni(), dfb = opt.m() * 2, nd = 0;
  if (2 * opt.pcmax() >= p) --dfb;
  vecdbl yvec(p);
  for (int j = 0; j < p; ++j)
  {
    if ((yvec[j] = cdata.fdata(j)) != BLANKf) ++nd;
  }
  if (nd <= dfb) return;        // 集計せず
  int d = opt.m() * 2 + 2;
  double w = 2.0 * M_PI / opt.t0();
  vecdbl tvec(d);
  tvec[0] = 1.0;
  for (int j = 0; j < p; ++j)
  {
    if (yvec[j] != BLANKf)
    {
      tvec[d - 1] = yvec[j];
      for (int l = 0; l < opt.m(); ++l)
      {
        int i = 2 * (l + 1),  k = opt.pcn(l);
        double tk = w * opt.t(j) * (double)k;
        tvec[i - 1] = cos(tk);
        tvec[i] = (2 * k < p) ? sin(tk) : 0.0;
      }
      for (int i = 0; i < d; ++i)
      {
        double x = tvec[i];
        for (int j = i; j < d; ++j) cosdc.smat_add(i, j, x * tvec[j]);
      }
    }
  }
  cosdc.nc(1);  cosdc.cal_sxy();
  cosdg += cosdc;
  cosdc.cal_pfun();
  cosdg.sr(cosdg.sr() + cosdc.sr());
}

void cosinorclass::put_dns(ofstream& fout, const cosdata& cosdc) const
// 振幅と位相をデータファイル形式で出力する
{
  int d = opt.m() * 2 + 2;
  datac cdata(d - 1);
  if (cosdc.nc() < 1)
  { for (int i = 0; i < d - 1; ++i) cdata.assign(i, BLANKs); }
  else
  {
    vecdbl tvec(d - 1);
    for (int i = 0; i < d - 1; ++i) tvec[i] = cosdc.smat(i, d - 1);
    pfun<double> pf(opt.t0(), opt.pcvec(), tvec);
    cdata.assign(0, pf.amp(0));
    for (size_t k = 1; k <= pf.m(); ++k)
    {
      int i = 2 * k;
      cdata.assign(i - 1, pf.amp(k));
      cdata.assign(i, pf.acr(k) / pf.w(k));
    }
  }
  cdata.put(fout);
}

void cosinorclass::put_cosinor(ostream& fout, const cosdata& cosd, bool groupc)
// コサイナー法結果の出力
{
  if (cosd.nc() < 1) return;    // 出力なし
  int d = opt.m() * 2 + 2, dfb = opt.m() * 2;
  if (2 * opt.pcmax() >= (int)(id.ni())) --dfb;
  vecdbl tvec(d - 1);
  for (int i = 0; i < d - 1; ++i) tvec[i] = cosd.smat(i, d - 1);
  pfun<double> pf(opt.t0(), opt.pcvec(), tvec);
  int dfr = cosd.n() - (dfb + 1) * cosd.nc();
  double sp = 0.0;
  if (groupc)                   // グループコサイナー法
  {
    putline(fout, LENline);
    fout << "ALL ";
    sp = cosd.syy() - cosd.smat(d - 1, d - 1);
  }
  else                          // シングルコサイナー法
  {
    fout << setw(4) << cosd.cn();
    sp = cosd.syy() - cosd.sr();
  }
  const double Seps = DBL_EPSILON * fmax(5.0, cosd.syy());
  if (sp < Seps) sp = 0.0;
  double r2 = (cosd.syy() > 0.0) ? sp / cosd.syy() : 0.0;
  double f = (cosd.sr() > 0.0) ? sp * (double)dfr / cosd.sr() / (double)dfb : 0.0;
  ldbl pr = ftop(f, dfb, dfr);
  fout << setw(10) << pf.amp(0) << setw(10) << pf.amp(1) << setw(10) << pf.acr(1) / pf.w(1);
  fout.setf(ios::fixed);
  fout << setw(10) << r2;
  fout.unsetf(ios::fixed);
  fout << setw(10) << f << setw(4) << dfb << setw(4) << dfr;
  fout.setf(ios::fixed);
  fout << setw(10) << pr;
  fout.unsetf(ios::fixed);
  fout << pind(pr) << "\n";
  if (groupc == false) return;
  if (opt.res("cc") && dfr > 0 && cosd.sr() > 0.0)
  {
    double vr = cosd.sr() / (double)dfr;
    double vx = vr * cosd.smat(1, 1), vy = vr * cosd.smat(2, 2);
    double vxy = vr * cosd.smat(1, 2);
    double a = 0.0, b = 0.0, h = 0.0;
    calelp(a, b, h, opt.cc(), dfb, dfr, vx, vy, vxy);
    fout << "\n" << opt.cc() * 100.0 << "% Simultaneous Confidence Ellipse of Amplitude and Acrophase\n";
    fout << "  Major Axis=" << a << "  Minor Axis=" << b << "  Angle of the Major=" << (M_PI * 0.5 - h) * 180.0 / M_PI << "degree\n";
  }
}

void cosinorclass::plot_ngpcos(const cosdata& cosdg)
// Ngraph用にコサイナーデータを出力する
{
  constexpr uint RDdef = 48;    // 極座標の分割数
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  plot_ngpid(fout);
  double dt = 2.0 * M_PI / (double)RDdef;
  int d = opt.m() * 2 + 2, dfb = opt.m() * 2;
  if (2 * opt.pcmax() >= (int)(id.ni())) --dfb;
  vecdbl tvec(d - 1);
  for (int i = 0; i < d - 1; ++i) tvec[i] = cosdg.smat(i, d - 1);
  pfun<double> pfg(opt.t0(), opt.pcvec(), tvec);
  int dfr = cosdg.n() - (dfb + 1) * cosdg.nc();
  double xmax = (opt.amax() == BLANKf) ? pfg.amp(1) * 1.5 : opt.amax();
  fout << "%<INNER CIRCUMFERENCE OF POLAR AXIS>\n";
  double t = 0.0;
  for (uint i = 0; i <= RDdef; ++i, t += dt)
  { fout << xmax * cos(t) << " " <<  xmax * sin(t) << "\n"; }
  fout << NGPdelm << "\n";
  fout << "%<OUTER CIRCUMFERENCE OF POLAR AXIS>\n";
  t = 0.0;
  for (uint i = 0; i <= RDdef; ++i, t += dt)
  { fout << 1.1 * xmax * cos(t) << " " <<  1.1 * xmax * sin(t) << "\n"; }
  fout << NGPdelm << "\n";
  t = 0.0;
  const uint Rdh = RDdef / 2;
  for (uint i = 0; i <= Rdh; ++i, t += 2.0 * dt)
  {
    fout << "= =\n";
    fout << "%<CIRCUMFERENCE SCALE:" << t << " RAD>\n";
    fout << xmax * cos(t) << " " <<  xmax * sin(t) << "\n";
    fout << 1.1 * xmax * cos(t) << " " <<  1.1 * xmax * sin(t) << "\n";
  }
  fout << NGPdelm << "\n";
  fout << "%<AMPLITUDE AND ACROPHASE(-s1 -dline -w80)>\n";
  fout << "0 0\n";
  double xm = cosdg.smat(2, d - 1), ym = cosdg.smat(1, d - 1);
  fout << xm << " " << ym << "\n";
  if (dfr < 1 || cosdg.sr() <= 0.0) return;
  double vr = cosdg.sr() / (double)dfr;
  double vx = vr * cosdg.smat(1, 1), vy = vr * cosdg.smat(2, 2);
  double vxy = vr * cosdg.smat(1, 2);
  double a = 0.0, b = 0.0, h = 0.0;
  calelp(a, b, h, opt.cc(), dfb, dfr, vy, vx, vxy);
  double x0 = 0.0, y0 = 0.0, x = 0.0, y = 0.0;
  fout << NGPdelm << "\n";
  fout << "%<CONFIDENCE ELLIPSE OF AMPLITUDE AND ACROPHASE(-s1 -dline -w40)>\n";
  t = 0.0;
  for (uint i = 0; i <= RDdef; ++i, t += dt)
  {
    x0 = a * cos(t);  y0 = b * sin(t);
    x = xm + x0 * cos(h) - y0 * sin(h);  y = ym + x0 * sin(h) + y0 * cos(h);
    fout << x << " " << y << "\n";
  }
}

void cosinorclass::plot_ngpid(ofstream& fout)
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "-s1 -dline";
  if (opt.amax() != BLANKf && opt.amax() > 0.0)
  {
    double xmax = opt.amax() * 1.1;
    fout << " -zx" << -xmax << "," << xmax << "," << xmax;
    fout << " -zy" << -xmax << "," << xmax << "," << xmax;
  }
  fout << "\n";
  fout << "%";  puttitle(fout, Title);  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%PERIOD ITEM  1:" << id.item(0) << "\n";
  fout << "%  :\n";
  fout << "%PERIOD ITEM" << setw(3) << id.ni() << "\n";
}

int cosinorclass::get(const char* c_fname) // データを読み込み累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void cosinorclass::put(int nc)  // 集計結果を出力する
{
  if (opt.res("res")) put_resid(cout);
  ofstream foutdns;
  if (opt.dnsname() != "")
  {
    if (ofopen(foutdns, (opt.dnsname()).c_str()) == false) return; // ファイルオープン失敗
    put_dnsid(foutdns);
  }
  if (nc < 1)                   // データなし
  {
    if (opt.res("res")) cout << "\n" << endl;
    return;
  }
  int p = id.ni();
  int q = p / 2;
  int d = opt.m() * 2 + 2, cn = 0;
  cosdata cosdg(0, d);
  if (opt.res("res") && opt.res("cos")) put_title(cout);
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    ++cn;
    cosdata cosdc(cn, d);
    cal_cos(cosdg, cosdc, *ditr);
    if (opt.res("res") && opt.res("cos")) put_cosinor(cout, cosdc, false);
    if (opt.dnsname() != "") put_dns(foutdns, cosdc);
  }
  cosdg.cal_gsxy();
  cosdg.cal_gpfun();
  if (opt.res("res"))
  {
    if (opt.res("cos")) put_cosinor(cout, cosdg, true);
    cout << "\n" << endl;
  }
  if (opt.ngpname() != "") plot_ngpcos(cosdg); // Ngraph用データ出力
}

int main(int argc, char* argv[]) // メインルーチン
{
  cosinorclass cosinor(argv[0]); // メインクラス変数
  try
  {
    cosinor.init();             // cosinorの初期設定
    bool fok = true;            // ファイル処理フラッグ
    int nc = 0;                 // 読み込んだケース数
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) nc += cosinor.get(nullptr); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { cosinor.setopt(argv[i] + 1);  fok = true; }
      else
      { nc += cosinor.get(argv[i]);  fok = false; }
    }
    cosinor.put(nc);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(cosinor.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(cosinor.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
