// fftce.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-11 10:40:58 tac>
//
// 高速フーリエ変換によるケース方向の周期解析を行う。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -W fftce fftce.C
//               -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "fftce  Version 7.41e         Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:fftc  [<OPTION> ...]  [<DATA FILE> ...]",
  "  The case based periodic analysis by Fast Fourier Transform.",
  "<OPTION>",
  "  -dh=<INTERVAL>:specify time interval",
  "  -f=<FILE名>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -k=<NUMBER>[,<NUMBER>,...]:specify the periodic component number to be",
  "    included in the periodic regression test(default:k=1,2)",
  "  -ngp={data|prc|psd|rsd}[,...]:select the data for Ngraph",
  "    data:raw data  prc:periodic regression curve",
  "    psd:power spectrum  rsd:contribution ratio spectrum",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=fftce_ngp.dat",
  "  -res=[-][{def|all|cc[<C-COEF>]|pctab[a]|stat}][,...]:select output results",
  "    -:wiht -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    cc[<C-COEF>]:confidence interval(0=<C=COEF<1)",
  "    pctab[a]:periodic component table(a:all periodic component)",
  "    stat:basic statistics",
  "  -T=<PERIOD>:specify the basic period(default:-T=24)"
};

const string Title = "=== Fourier Analysis ==="; // 表題
const string NGPNAMEdef = "fftce_ngp.dat"; // 省略時Ngraphファイル名
constexpr double T0def = 24.0;    // 省略時基本周期

using mapsb = map<string, bool>; // boolマップの別名定義
using mapsb_itr = mapsb::iterator;        // 反復子の別名定義
using mapsb_citr = mapsb::const_iterator; // const反復子の別名定義

using lstdbl = list<double>;    // 倍精度実数リストの別名定義


enum pctabtype                  // 周期成分表出力の種類
{
  pctabnon,                     // 出力せず
  pctabauto,                    // 選択した周期成分だけ出力
  pctaball,                     // 全て出力
  pctabetc                      // その他
};

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  double vdh;                   // データの時間間隔
  lstr klst;                    // 周期成分番号リスト
  vecint vpcvec;                // 周期成分番号配列
  string vngpname;              // Ngraphファイル基本名
  lstr vngplst;                 // Ngraph出力リスト
  mapsb vngpmap;                // Ngraph出力の有無マップ
  pctabtype vpctab;             // 周期成分表の出力フラッグ
  lstr vreslst;                 // 分析結果出力リスト
  mapsb vresmap;                // 分析結果出力の有無マップ
  double vt0;                   // 基本周期
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  double dh() const             // データの時間間隔
  { return vdh; }
  int m() const                 // 周期成分数を返す
  { return vpcvec.size(); }
  int pcn(size_t k) const       // 周期成分番号を返す
  { return (k < vpcvec.size()) ? vpcvec[k] : 0; }
  pctabtype pctab() const       // 周期成分表出力フラッグを返す
  { return vpctab; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 分析結果出力フラッグを返す
  double t0() const             // 基本周期を返す
  { return vt0; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(int p);             // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vdh(1.0), klst(), vpcvec(), vngpname(), vngplst(), vngpmap(), vpctab(pctabauto), vreslst(), vresmap(), vt0(T0def)
// コンストラクタ
{
  if (c_pname == nullptr) return;     // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  mapsb_itr ngpitr = vngpmap.find(code);
  if (ngpitr == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  mapsb_itr resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "dh")           // データの時間間隔の指定
  {
    vdh = strtodf(opt[1]);
    if (vdh == BLANKf || vdh <= 0.0) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "k")       // 周期成分番号の指定
  {
    if (strsplit(klst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "T")       // 基本周期の指定
  {
    vt0 = strtodf(opt[1]);
    if (vt0 == BLANKf || vt0 <= 0.0) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(int p)        // オプションのチェックとセット
{
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  int q = p / 2;                // 周期成分最大数
  if (klst.empty())             // 周期成分指定がない時
  {
    vpcvec.resize(2);
    vpcvec[0] = 1;  vpcvec[1] = 2;
  }
  else if (setpcvec(vpcvec, klst, q) < 0)
  { return disperr(vpname, "Wrong periodic component !"); }
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["data"] = vngpmap["prc"] = vngpmap["psd"] = vngpmap["rsd"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["data"] = vngpmap["prc"] = vngpmap["psd"] = vngpmap["rsd"] = false;
    for (string ngp : vngplst)
    {
      mapsb_itr ngpitr = vngpmap.find(ngp);
      if (ngpitr == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["pctab"] = vresmap["stat"] = true;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["cc"] = vresmap["pctab"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      mapsb_itr resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["pctab"] = vresmap["stat"] = true;
      }
      else if (res == "all")   // 全て出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["pctab"] = vresmap["stat"] = true;
        vpctab = pctaball;      // 全ての周期成分を出力する
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (res == "pctaba")
      { vresmap["pctab"] = true;  vpctab = pctaball; }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class varclass                  // 変数クラス
{
private:
  lstdbl vylst;                 // 測定データのリスト
  int vnp;                      // データ数
  double vmean;                 // 平均値
  double vst;                   // 平方和
  pfun<double> vpf0;            // 全体周期関数オブジェクト
  pfun<double> vpf;             // 選択周期関数オブジェクト
  void cal_fftc(vecdbl& abvec, const option& opt); // 基礎統計の計算とFFT
public:
  varclass() : vylst(), vnp(0), vmean(0.0), vst(0.0), vpf0(), vpf() // コンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  const lstdbl& ylst() const    // 測定データのリストを返す
  { return vylst; }
  int np() const                // データ数を返す
  { return vnp; }
  double mean() const           // 平均値を返す
  { return vmean; }
  double st() const             // 平方和を返す
  { return vst; }
  const pfun<double> pf0() const // 全体周期関数オブジェクトを返す
  { return vpf0; }
  const pfun<double> pf() const // 選択周期関数オブジェクトを返す
  { return vpf; }
  int cal(option& opt, ldata& datalst); // データを集計してFFT計算をする
};

void varclass::cal_fftc(vecdbl& abvec, const option& opt)
// 基礎統計の計算とFFT
{
  int p = vnp;
  int q = p / 2;
  const double wt = opt.dh() * 2.0 * M_PI / (double)p;
  abvec.assign(p + 2, 0.0);
  int j = 0;
  for (double y : vylst)
  {
    for (int k = 1; k <= q; ++k)
    {
      int i = 2 * k;
      double tk = wt * (double)(k * j);
      abvec[i - 1] += y * cos(tk);
      if (i < p) abvec[i] += y * sin(tk);
    }
    abvec[0] += y;  abvec[p + 1] += y * y;
    ++j;
  }
  for (int k = 1; k <= q; ++k)
  {
    int i = 2 * k;
    if (i == p) abvec[i - 1] /= (double)p;
    else { abvec[i - 1] /= (double)p * 0.5;  abvec[i] /= (double)p * 0.5; }
  }
  abvec[p + 1] -= abvec[0] * abvec[0] / (double)p; // データの平方和(St)
  const double ABeps = DBL_EPSILON * fmax(5.0, wt);
  if (abvec[p + 1] < ABeps) abvec[p + 1] = 0.0;
  abvec[0] /= (double)p;        // データの平均(My)
}

int varclass::cal(option& opt, ldata& datalst) // データを集計してFFT計算をする
{
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    double y = ditr->fdata(0);
    if (y != BLANKf) vylst.push_back(y);
  }
  if (opt.check(vylst.size()) != 0) return disperr(opt.pname(), "実行時オプションエラー！");
  vnp = vylst.size();
  int p = vnp;
  vecdbl abvec0(p + 2);
  cal_fftc(abvec0, opt);
  vmean = abvec0[0];  vst = abvec0[p + 1];
  int q = p / 2;
  vecint pcvec0(q, 0);
  for (int k = 0; k < q; ++k) pcvec0[k] = k + 1;
  pfun pf0(opt.t0(), pcvec0, abvec0);
  pcvec0.resize(opt.m());
  vecdbl abvec(opt.m() * 2 + 2);
  abvec[0] = abvec0[0];
  for (int k = 0; k < opt.m(); ++k)
  {
    pcvec0[k] = opt.pcn(k);
    int i = (k + 1) * 2, j = opt.pcn(k) * 2;
    abvec[i - 1] = abvec0[j - 1];
    abvec[i] = abvec0[j];
  }
  abvec[opt.m() * 2 + 1] = abvec0[p + 1];
  pfun pf(opt.t0(), pcvec0, abvec);
  vpf0 = pf0;  vpf = pf;
  return q;                     // 周期成分数を返す
}

class fftcclass                 // fftcメインクラス
{
private:
  option opt;                   // オプション
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  void put_resid(ostream& fout); // 解析結果のタイトルとデータ識別情報を出力する
  void put_stat(ostream& fout, const varclass& var);
  // 時期別基礎統計量を出力する
  void put_fftcfun(ostream& fout, int p, const pfun<double>& pf, double st) const;
  // FFT用周期関数パラメターを出力する
  void put_fftctab(ostream& fout, int p, const pfun<double>& pf, double st);
  // フーリエ解析用分散分析表を出力する
  void put_fftctab0(ostream& fout, int p, const pfun<double>& pf, double sp, int dfp, double sr, int dfr, double st, int dft) const;
  // FFT用周期回帰分散分析表を出力する
  void plot_ngp(const varclass& var); // Ngraph用データを出力する
  void plot_ngpid(ofstream& fout);
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_ngpdata(ofstream& fout, const lstdbl& ylst);
  // Ngraph用に生データを出力する
  void plot_ngpprc(ofstream& fout, const pfun<double>& pf) const;
  // Ngraph用に周期回帰曲線データを出力する
  void plot_ngppsd(ofstream& fout, int p, const pfun<double>& pf) const;
  // Ngraph用にパワースペクトルデータを出力する
  void plot_ngprsd(ofstream& fout, int p, const pfun<double>& pf, double st) const;
  // Ngraph用に寄与率スペクトルデータを出力する
public:
  fftcclass(const char* c_pname = nullptr) : opt(c_pname), fname(), id(), datalst()
  // コンストラクタ
  {}
  ~fftcclass()                  // デストラクタ
  {}
  const string& pname() const   // プログラム名を返す
  { return opt.pname(); }
  void init()      // 初期設定を行う(コンストラクタ中ではできない処理)
  { opt.setdef(); }
  void setopt(const string& optstr, char devc = OPTdev) // オプションを設定する
  { opt.set(optstr, devc); }
  int get(const char* c_fname); // データを読み込み累積する
  void put(int nc);             // 集計結果を出力する
};

void fftcclass::put_resid(ostream& fout)
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "DATA ITEM:" << id.item(0) << "\n";
}

void fftcclass::put_stat(ostream& fout, const varclass& var)
// データの基礎統計量を出力する
{
  int p = var.np();
  double vr = (p > 1) ? var.st() / (p - 1) : 0.0;
  fout << "\nINTERVAL:Basic Statistics\n";
  fout << setw(8) << opt.dh();
  fout << ":";
  putmsd(fout, p, var.mean(), vr);
}

void fftcclass::put_fftcfun(ostream& fout, int p, const pfun<double>& pf, double st) const
// FFT用周期関数パラメターを出力する
{
  fout << "\n<PERIODIC REGRESSION EQUATION>\n";
  fout << "y = A0 + sum{Ak x cos(wk x t - hk)}\n";
  fout << "k=1,...,p:periodic component  A0:MESOR  Ak:amplitude\n";
  fout << "wk=360deg x k/T:Angular Frequency  hk:Acrophase  T=" << pf.t(0) << ":Basic Period\n";
  fout << "T/k:Period  f=1/T:Basic Frequency  fk=k/T:Frequency\n";
  if (opt.pctab() == pctabnon) return;
  fout << "   k            Ak        hk     hk/wk          ak          bk     r^2     T/k\n";
  constexpr int LENline = 78;
  putline(fout, LENline);
  fout << "   0" << setw(14) << pf.amp(0) << "(MESOR)\n";
  double deg = 180.0 / M_PI;
  for (size_t k = 1; k <= pf.m(); ++k)
  {
    double dfk = (2 * pf.pcn(k) < p) ? 2.0 : 1.0;
    double r2 = pf.amp(k) * pf.amp(k) * (double)p / dfk / st;
    fout << setw(4) << pf.pcn(k) << setw(14) << pf.amp(k);
    fout << setw(10) << pf.acr(k) * deg;
    fout << setw(10) << pf.acr(k) / pf.w(k);
    fout << setw(12) << pf.a(k) << setw(12) << pf.b(k);
    int prec0 = fout.precision(5); // 小数点以下の精度を変更する
    fout.setf(ios::fixed);
    fout << setw(8) << r2;
    fout.unsetf(ios::fixed);
    fout << setw(8) << pf.t(k) << setprecision(prec0) << "\n";
  }
  putline(fout, LENline);
}

void fftcclass::put_fftctab(ostream& fout, int p, const pfun<double>& pf, double st)
// フーリエ解析用分散分析表を出力する
{
  if (st <= 0.0) return;
  int dft = p - 1;
  double sp = 0.0;
  int dfp = 0;
  for (size_t k = 1; k <= pf.m(); ++k)
  {
    int dfk = (2 * pf.pcn(k) < p) ? 2 : 1;
    dfp += dfk;
    sp += pf.amp(k) * pf.amp(k) * (double)p / (double)dfk;
  }
  const double Seps = DBL_EPSILON * fmax(5.0, st);
  if (sp < Seps) sp = 0.0;
  double sr = st - sp;
  if (sr < Seps) sr = 0.0;
  int dfr = dft - dfp;
  double v = (dfr > 0) ? sr / (double)dfr / (double)(dft + 1) : 0.0;
  double ci = calmcci(mcfis, opt.cc(), 1, dfp, dfr, 1, v);
  if (opt.res("cc"))
  {
    fout << opt.cc() * 100.0 << "% CI of MESOR:";
    fout << pf.amp(0) << " +/- " << ci << "(" << pf.amp(0) - ci << " - " << pf.amp(0) + ci << ")\n";
  }
  put_fftctab0(fout, p, pf, sp, dfp, sr, dfr, st, dft);
}

void fftcclass::put_fftctab0(ostream& fout, int p, const pfun<double>& pf, double sp, int dfp, double sr, int dfr, double st, int dft) const
// FFT用周期回帰分散分析表を出力する
{
  constexpr int LENline = 78;
  string tstr("Periodic ANOVA Table");
  fout << "\n" << setw((LENline + tstr.size()) / 2) << tstr;
  fout << "\nFACTOR    Sum of Squares        DF  Mean Squares     F-value       p-value\n";
  putline(fout, LENline);
  double v = 0.0, f = 0.0, pr = 0.0;
  for (size_t k = 1; k <= pf.m(); ++k)
  {
    int dfk = (2 * pf.pcn(k) < p) ? 2 : 1;
    double spk = pf.amp(k) * pf.amp(k) * (double)(dft + 1) / (double)dfk;
    calvfp(v, f, pr, spk, dfk, sr, dfr);
    fout << "PERIOD" << setw(4) << pf.pcn(k) << setw(14) << spk << setw(10) << dfk;
    fout << setw(14) << v << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
  }
  putline(fout, LENline);
  double vr = calvfp(v, f, pr, sp, dfp, sr, dfr);
  tstr = "ALL PERIOD";
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sp;
  fout << setw(10) << dfp << setw(14) << v << setw(12) << f;
  fout << setw(14) << pr << pind(pr) << "\n";
  tstr = "RESIDUALS";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sr;
  fout << setw(10) << dfr << setw(14) << vr << "\n";
  putline(fout, LENline);
  tstr = "TOTAL";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << st;
  fout << setw(10) << dft << endl;
}

void fftcclass::plot_ngp(const varclass& var)
// Ngraph用データを出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("data"))          // 生データ
  {
    fout << "-s1 -dmark,00 -o300\n";
    plot_ngpid(fout);
    plot_ngpdata(fout, var.ylst());
    delml = true;
  }
  if (opt.ngp("prc"))           // 周期回帰曲線
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -c255,0,0\n";
      delml = true;
    }
    plot_ngpid(fout);
    plot_ngpprc(fout, var.pf());
  }
  if (opt.ngp("psd"))           // パワースペクトル
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -c255,0,0\n";
      delml = true;
    }
    plot_ngpid(fout);
    plot_ngppsd(fout, var.np(), var.pf0());
  }
  if (opt.ngp("rsd"))           // 寄与率スペクトル
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dline -c255,0,0\n";
    plot_ngpid(fout);
    plot_ngprsd(fout, var.np(), var.pf0(), var.st());
  }
}

void fftcclass::plot_ngpid(ofstream& fout)
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%DATA ITEM:" << id.item(0) << "\n";
}

void fftcclass::plot_ngpdata(ofstream& fout, const lstdbl& ylst)
// Ngraph用に個々のデータを出力する
{
  fout << "%<RAW DATA>\n";
  double t = 0.0;
  for (double y : ylst)
  {
    fout << t << " " << y << endl;
    t += opt.dh();
  }
}

void fftcclass::plot_ngpprc(ofstream& fout, const pfun<double>& pf) const
// Ngraph用周期回帰曲線を出力する
{
  fout << "%<PERIODIC REGRESSION CURVE>\n";
  constexpr uint RDdef = 60;    // 曲線の分解数
  double dt = pf.t(0) / (double)RDdef;
  double t = 0.0;
  for (uint i = 0; i <= RDdef; ++i, t += dt)
  { fout << t << " " << pf.fun(t) << endl; }
}

void fftcclass::plot_ngppsd(ofstream& fout, int p, const pfun<double>& pf) const
// Ngraph用にパワースペクトルデータを出力する
{
  fout << "%<POWER SPECTRUM>\n";
  double ff = 1.0 / opt.t0();   // 周期成分番号->周波数変換ファクター
  for (size_t k = 1; k <= pf.m(); ++k)
  {
    double dfk = (2 * pf.pcn(k) < p) ? 2.0 : 1.0;
    double apow = pf.amp(k) * pf.amp(k) / dfk;
    fout << (double)k * ff << " " << apow << endl;
  }
}

void fftcclass::plot_ngprsd(ofstream& fout, int p, const pfun<double>& pf, double st) const
// Ngraph用に寄与率スペクトルデータを出力する
{
  fout << "%<CONTRIBUTION RATIO SPECTRUM>\n";
  double ff = 1.0 / opt.t0();   // 周期成分番号->周波数変換ファクター
  for (size_t k = 1; k <= pf.m(); ++k)
  {
    double dfk = (2 * pf.pcn(k) < p) ? 2.0 : 1.0;
    double r2 = pf.amp(k) * pf.amp(k) * (double)p / dfk / st;
    fout << (double)k * ff << " " << r2 << endl;
  }
}

int fftcclass::get(const char* c_fname) // データを読み込み累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (id.ni() > 1) throw runtime_error("Number of item must be 1 !");
  return nc;
}

void fftcclass::put(int nc)     // 集計結果を出力する
{
  if (nc < 1)                   // データ無し
  {
    if (opt.res("res"))
    {
      put_resid(cout);
      cout << "\n" << endl;
    }
    return;
  }
  varclass var;                 // 変数オブジェクト
  if (var.cal(opt, datalst) < 0) return;
  if (opt.res("res"))
  {
    put_resid(cout);
    if (opt.res("stat")) put_stat(cout, var);
    if (opt.res("pctab"))
    {
      if (opt.pctab() == pctaball) put_fftcfun(cout, var.np(), var.pf0(), var.st());
      else put_fftcfun(cout, var.np(), var.pf(), var.st());
      put_fftctab(cout, var.np(), var.pf(), var.st());
    }
    cout << "\n" << endl;
  }
  if (opt.ngpname() != "") plot_ngp(var);
}

int main(int argc, char* argv[]) // メインルーチン
{
  fftcclass fftc(argv[0]);      // メインクラス変数
  try
  {
    fftc.init();                // fftcの初期設定
    bool fok = true;            // ファイル処理モードフラッグ
    int nc = 0;                 // 読み込んだケース数
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) nc += fftc.get(nullptr); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { fftc.setopt(argv[i] + 1);  fok = true; }
      else
      { nc += fftc.get(argv[i]);  fok = false; }
    }
    fftc.put(nc);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(fftc.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(fftc.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
