// mvae.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-11 11:02:27 tac>
//
// 移動平均を計算する。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -W mvae mvae.C
//               -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "mvae  Version 7.41e          Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:mvae  [<OPTION> ...]  [<DATA FILE> …]",
  "  Computing moving averages.",
  "<OPTION>",
  "  -dnsname[=<FILE>]:specify the file to output DANS data",
  "    default:no output  -dnsname:-dnsname=mvae_dns.d",
  "  -dt=<INTERVAL>:specify time interval(default:-dt=1)",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -k=<NUMBER>:specify the number of periods for moving averages(NUMBER>=1)",
  "    compute the average of k before and after(2k+1 in total)",
  "    default:-k=1",
  "  -ngp={mse|mva}[,...]:select the data for Ngraph",
  "    mse:mean +/- SE  mva:moving averages  default:all",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=mvae_ngp.dat",
  "  -p=<ORDER>:specify the order of the regression equation for moving",
  "    averages(1=<OEDER<=2k)  default:-p=1",
  "  -res[={-|+}]:select output results(-res:-res=+)",
  "    -:no output  +:output(default)",
  "  -t=<TIME1>,<TIME2>[,<TIME3>,...]:specify the relative time of items",
  "    default:-t=0,1,...,(number of items - 1)"
};

const string Title = "=== Moving Average ==="; // 表題
const string DNSNAMEdef = "mvae_dns.d"; // 省略時DANSファイル名
const string NGPNAMEdef = "mvae_ngp.dat"; // 省略時Ngraphファイル名

enum ngptype                    // Ngraph用出力データの種類
{
  ngpnon,                       // なし
  ngpdns,                       // データファイル形式の移動平均
  ngpmse,                       // 平均値±SE
  ngpmva,                       // 移動平均値
  ngpetc                        // その他
};

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  string vdnsname;              // DANSファイル名
  double vdt;                   // 時期間隔
  int vk;                       // 移動平均を計算する前後の時期数
  string vngpname;              // Ngraphファイル名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  bool vres;                    // 解析結果出力フラッグ
  int vp;                       // 回帰式の次数
  lstr tlst;                    // 項目時間リスト
  vecdbl vtvec;                 // 項目時間ベクトル
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  const string& dnsname() const // DANSデータファイル名を返す
  { return vdnsname; }
  double dt() const             // 時期間隔を返す
  { return vdt; }
  int k() const                // 移動平均を計算する前後の時期数を返す
  { return vk; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  int p() const                 // 回帰式の次数を返す
  { return vp; }
  bool res() const              // 解析結果出力フラッグを返す
  { return vres; }
  double t(size_t idx) const    // 項目時間を返す
  { return (idx < vtvec.size()) ? vtvec[idx] : BLANKf; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vk(1), vdt(1.0), vdnsname(), vngpname(), vngplst(), vngpmap(), vp(1), vres(true), tlst(), vtvec()
// コンストラクタ
{
  if (c_pname == nullptr) return;     // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngppos = vngpmap.find(code);
  if (ngppos == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "dnsname")      // DANSファイル名指定
  {
    if (opt[1].empty()) vdnsname = DNSNAMEdef;
    else vdnsname = opt[1];
    if (vdnsname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "dt")      // 時期間隔の指定
  {
    vdt = strtodf(opt[1]);
    if (vdt <= 0.0 || vdt == BLANKf) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "k")       // 前後の時期数指定
  {
    char* ptr = nullptr;
    vk = strtol(opt[1].c_str(), &ptr, 0);
    if (vk < 1) err = badopt(vpname, optstr, USAGE, 1.0, (double)(INT_MAX / 2));
  }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "p")       // 回帰式の次数指定
  {
    char* ptr = nullptr;
    vp = strtol(opt[1].c_str(), &ptr, 0);
    if (vp < 1 || 2 * vk < vp) err = badopt(vpname, optstr, USAGE, 1.0, 2.0 * (double)vk);
  }
  else if (opt[0] == "res")     // 解析結果出力フラッグの指定
  {
    if (opt[1].empty() || opt[1] == "+") vres = true;
    else if (opt[1] == "-") vres = false;
  }
  else if (opt[0] == "t")       // 項目時間の指定
  {
    if (strsplit(tlst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションをチェックする
{
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (ni < 2 * (size_t)vk + 1) return disperr(vpname, "Not enough items !");
  vtvec.resize(ni);
  for (size_t j = 0; j < ni; ++j) vtvec[j] = (double)j * vdt;
  if (settvec(vtvec, tlst, ni) < 0) return disperr(vpname, "Wrong time of item !");
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["mse"] = vngpmap["mva"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["mse"] = vngpmap["mva"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngppos = vngpmap.find(ngp);
      if (ngppos == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

class varclass                  // 変数クラス
{
private:
  vector<statv<double>> statvec; // 実測値集計用基礎統計量ベクトル
  statv<double> mstat;           // 移動平均計算用基礎統計量
  mat<double> smat;              // 回帰係数計算用行列
public:
  varclass(size_t ni) : statvec(ni), mstat(), smat() // コンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int n() const                 // 平均値数を返す
  { return mstat.n(); }
  double mean() const           // 平均値の平均値を返す
  { return mstat.mean(); }
  int n(size_t idx) const       // 時期別例数を返す
  { return (idx < statvec.size()) ? statvec[idx].n() : 0; }
  double mean(size_t idx) const // 時期別平均値を返す
  { return (idx < statvec.size()) ? statvec[idx].mean() : 0; }
  double se(size_t idx) const   // 時期別標準誤差を返す
  { return (idx < statvec.size()) ? statvec[idx].se() : 0; }
  int cal(option& opt, ldata& datalst); // データを集計して統計量を求める
  size_t regc(const option& opt, size_t idx, size_t q);
  double fmva(double t) const;          // 移動平均を返す
};

int varclass::cal(option& opt, ldata& datalst) // データを集計して統計量を求める
{
  for (size_t j = 0; j < statvec.size(); ++j) statvec[j].clear();
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    for (size_t j = 0; j < statvec.size(); ++j)
    {
      if (ditr->fdata(j) != BLANKf) statvec[j].sigma(ditr->fdata(j));
    }
  }
  for (size_t j = 0; j < statvec.size(); j++)
  {
    if (statvec[j].n() > 0) mstat.sigma(statvec[j].mean());
  }
  smat.resize(opt.p() + 2, opt.p() + 2);
  return mstat.n();             // 平均値数を返す
}

size_t varclass::regc(const option& opt, size_t idx, size_t q)
// 回帰式を計算する
{
  size_t p = smat.r() - 2;
  smat = 0.0;
  vecdbl tvec(smat.r());
  tvec[0] = 1.0;
  size_t k = 0, sidx = 0;
  while (k < q && idx < statvec.size())
  {
    if (statvec[idx].n() > 0)
    {
      if (k == 0) sidx = idx;
      double ti = opt.t(idx);
      for (size_t j = 1; j <= p; ++j)
      {
        tvec[j] = ti;
        ti *= tvec[j];
      }
      tvec[p + 1] = statvec[idx].mean();
      for (size_t i = 0; i <= p + 1; ++i)
      {
        double x = tvec[i];
        for (size_t j = i; j <= p + 1; ++j) smat.add(i, j, x * tvec[j]);
      }
      ++k;
    }
    ++idx;
  }
  for (size_t i = 0; i < smat.r(); ++i)
  { for (size_t j = i + 1; j < smat.r(); ++j) smat.assign(j, i, smat(i, j)); }
  for (size_t i = 0; i <= p; ++i) smat.sweep(i); // 回帰係数を計算する
  return sidx;               // データがあった最初のインデックスを返す
}

double varclass::fmva(double t) const // 移動平均を返す
{
  size_t yidx = smat.r() - 1;
  double y = smat(0, yidx), x = t;
  for (size_t i = 1; i < yidx; ++i)
  {
    y += smat(i, yidx) * x;
    x *= t;
  }
  return y;                     // 回帰推定値を移動平均として返す
}

class mvaclass                  // mvaメインクラス
{
private:
  option opt;                   // オプション
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  void put_resid(ostream& fout); // 解析結果のタイトルとデータ識別情報を出力する
  void put_mva(ostream& fout, varclass& var); // 移動平均表を出力する
  void put_dnsmva(varclass& var); // 移動平均をデータファイル形式で出力する
  void plot_ngp(varclass& var);   // Ngraph用データを出力する
  void plot_ngpid(ofstream& fout);
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_ngpmva(ofstream& fout, ngptype ngp, varclass& var);
  // Ngraph用に移動平均/平均値±SEを出力
public:
  mvaclass(const char* c_pname = nullptr) : opt(c_pname), fname(), id(), datalst()
  // コンストラクタ
  {}
  ~mvaclass()                   // デストラクタ
  {}
  const string& pname() const   // プログラム名を返す
  { return opt.pname(); }
  void init()      // 初期設定を行う(コンストラクタ中ではできない処理)
  { opt.setdef(); }
  void setopt(const string& optstr, char devc = OPTdev) // オプションを設定する
  { opt.set(optstr, devc); }
  int get(const char* c_fname); // データを読み込み累積する
  void put(int nc);             // 集計結果を出力する
};

void mvaclass::put_resid(ostream& fout)
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "PERIOD ITEM  1:" << id.item(0) << "\n";
  fout << "  :\n";
  fout << "PERIOD ITEM" << setw(3) << id.ni();
  fout << ":" << id.item(id.ni() - 1) << "\n";
}

void mvaclass::put_mva(ostream& fout, varclass& var)
// 移動平均表を出力する
{
  int q = 2 * opt.k() + 1;
  fout << "\n<DATA AND MOVING AVERAGES>\n";
  fout << "k=" << opt.k() << "(n of datas=" << q;
  fout << ")  Order of Partial Regression=" << opt.p() << "\n";
  constexpr int LENline = 72;
  putline(fout, LENline);
  int lidx = opt.k(), hidx = var.n() - 1 - opt.k();
  int sidx = var.regc(opt, 0, q);
  statv<double> mvatotal;
  for (int i = 0, j = 0; i < var.n(); j++)
  {
    if (var.n(j) > 0)
    {
      if (lidx < i && i <= hidx) sidx = var.regc(opt, sidx + 1, q);
      double mva = var.fmva(opt.t(j));
      ++i;
      fout << "t=" << setw(6) << opt.t(j);
      fout << left << ":data=" << setw(14) << var.mean(j);
      fout << "moving average=" << setw(14) << mva;
      fout << "diff=" << setw(14) << var.mean(j) - mva << right << "\n";
      mvatotal.sigma(mva);
    }
  }
  putline(fout, LENline);
  fout << left << "MEAN    :DATA=" << setw(14) << var.mean();
  fout << "MOVING AVERAGE=" << setw(14) << mvatotal.mean();
  fout << "DIFF=" << setw(14) << var.mean() - mvatotal.mean() << right << endl;
}

void mvaclass::put_dnsmva(varclass& var)
// 移動平均をデータファイル形式で出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.dnsname()).c_str()) == false) return; // ファイルオープン失敗
  id.put(fout, formtag);
  int q = 2 * opt.k() + 1;
  datac cdata(id.ni());
  int lidx = opt.k(), hidx = var.n() - 1 - opt.k();
  int sidx = var.regc(opt, 0, q);
  statv<double> mvatotal;
  for (int i = 0, j = 0; i < var.n(); i++,j++)
  {
    double mva = BLANKf;
    if (var.n(j) > 0)
    {
      if (lidx < i && i <= hidx) sidx = var.regc(opt, sidx + 1, q);
      mva = var.fmva(opt.t(j));
    }
    cdata.assign(i, mva);
  }
  cdata.put(fout);
}

void mvaclass::plot_ngp(varclass& var) // Ngraph用データを出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("mva"))           // 移動平均
  {
    fout << "-s1 -dline -w80\n";
    plot_ngpid(fout);
    plot_ngpmva(fout, ngpmva, var);
    delml = true;
  }
  if (opt.ngp("mse"))           // 平均値±SE
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dline -w80\n";
    plot_ngpid(fout);
    plot_ngpmva(fout, ngpmse, var);
  }
}

void mvaclass::plot_ngpid(ofstream& fout)
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%PERIOD ITEM  1:" << id.item(0) << "\n";
  fout << "%  :\n";
  fout << "%PERIOD ITEM" << setw(3) << id.ni();
  fout << ":" << id.item(id.ni() - 1) << "\n";
}

void mvaclass::plot_ngpmva(ofstream& fout, ngptype ngp, varclass& var)
// Ngraph用に移動平均/平均値±SEを出力
{
  int q = 2 * opt.k() + 1;
  if (ngp == ngpmse) fout << "%Mean(-dmark,00 -o300) +/- SE(-derrorbar_y -w40 -o500)\n";
  else fout << "%Moving Averages:k=" << opt.k() << "(n of partial datas=" << q << ")  Order of Partial Regression=" << opt.p() << "\n";
  int lidx = opt.k(), hidx = var.n() - 1 - opt.k(), sidx = var.regc(opt, 0, q);
  for (int i = 0, j = 0; i < var.n(); ++j)
  {
    if (var.n(j) > 0)
    {
      if (lidx < i && i <= hidx) sidx = var.regc(opt, sidx + 1, q);
      double mva = var.fmva(opt.t(j));
      double se = var.se(j);
      ++i;
      fout << opt.t(j) << " ";
      if (ngp == ngpmse)
      { fout << var.mean(j) << " " << -se << " " << se << "\n"; }
      else
      { fout << mva << endl; }
    }
  }
}

int mvaclass::get(const char* c_fname) // データを読み込み累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void mvaclass::put(int nc)      // 集計結果を出力する
{
  if (nc < 1)                   // データ無し
  {
    if (opt.res())
    {
      put_resid(cout);
      cout << "\n" << endl;
    }
    return;
  }
  varclass var(id.ni());        // 変数オブジェクト
  if (var.cal(opt, datalst) < opt.k() * 2 + 1) return;
  if (opt.res())
  {
    put_resid(cout);
    put_mva(cout, var);
    cout << "\n" << endl;
  }
  if (opt.dnsname() != "") put_dnsmva(var);
  if (opt.ngpname() != "") plot_ngp(var);
}

int main(int argc, char* argv[]) // メインルーチン
{
  mvaclass mva(argv[0]);        // メインクラス変数
  try
  {
    mva.init();                 // mvaの初期設定
    bool fok = true;            // ファイル処理フラッグ
    int nc = 0;                 // 読み込んだケース数
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) nc += mva.get(nullptr); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { mva.setopt(argv[i] + 1);  fok = true; }
      else
      { nc += mva.get(argv[i]);  fok = false; }
    }
    mva.put(nc);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(mva.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(mva.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
