// tcovae.C Version 7.42e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-11 13:24:32 tac>
//
// 最初の項目を群項目、以後の項目を時期項目とした二元配置型時系列共分散分析。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -W tcovae tcovae.C
//               -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "tcovae  Version 7.42e        Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:tcova [<OPTION> ...]  [<DATA FILE> ...]",
  "  The time series ANCOVA with the first item as group and subsequent items",
  "  as periods.",
  "<OPTION>",
  "  -am={0|1|2}:select the method of ANCOVA",
  "    0:simple method using only means  1:one-way layout method",
  "    2:two-way layout method(default)",
  "  -dt=<INTERVAL>:specify time interval(default:-dt=1)",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -n=<ORDER>:specify the oder of regression function(1=<OEDER)",
  "  -res=[-][{def|all|atab|cc[<C-COEF>]|m[{a|d}][{b|f|s|t}]|statg|statt}]",
  "      [,...]:select output results",
  "    -:wiht -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    atab:ANCOVA table",
  "    cc[<C-COEF>]:confidence interval(0=<C=COEF<1  cc:cc0.95)",
  "    m[{a|d}][{b|f|s|t}]:multiple comparison(ma:group mean  md:non-parallelism)",
  "      b:Bonferroni type  f:Fisher type  s:Scheffe type  t:Tukey type",
  "    statg:group statistics  statt:overall statistics",
  "  -t=<TIME1>,<TIME2>[,<TIME3>,...]:specify the relative time of items",
  "    default:-t=0,1,2,…,(number of items - 1)"
};

const string Title = "=== Time Series ANCOVA ==="; // 表題

enum amtype                     // 周期解析の種類
{
  amnon,                        // なし
  am0,                          // 平均値を用いる単純法
  am1,                          // 一元配置法
  am2,                          // 二元配置法
  ametc                         // その他
};

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  amtype vam;                   // 分散分析の種類
  double vcc;                   // 信頼係数
  double vdt;                   // 時期間隔
  mctype vma;                   // 修正群差の多重比較法
  mctype vmd;                   // 非平行性の多重比較法
  int vn;                       // 共変数の次数
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  lstr tlst;                    // 共変数の値リスト
  vecdbl vtvec;                 // 共変数の値ベクトル
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  amtype am() const             // 分散分析の種類を返す
  { return vam; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  double dt() const             // 時期間隔を返す
  { return vdt; }
  mctype ma() const             // 修正群差の多重比較法を返す
  { return vma; }
  mctype md() const             // 非平行性の多重比較法を返す
  { return vmd; }
  int n() const                 // 共変数の次数
  { return vn; }
  bool res(const string& code); // 分析結果出力フラッグを返す
  double t(size_t idx) const    // 時期項目の値を返す
  { return (idx < vtvec.size()) ? vtvec[idx] : 0.0; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(int ni);            // オプションをチェックする
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vam(am2), vcc(CCdef), vdt(1.0), vma(mctuk), vmd(mcbon), vn(1), tlst(), vtvec()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "am")           // 分散分析の手法指定
  {
    if (opt[1].empty() || opt[1] == "2") vam = am2;
    else if (opt[1] == "-") vam = amnon;
    else if (opt[1] == "0") vam = am0;
    else if (opt[1] == "1") vam = am1;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "dt")      // 時期間隔の指定
  {
    vdt = strtodf(opt[1]);
    if (vdt <= 0.0 || vdt == BLANKf) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "n")       // 共変数の次数指定
  {
    char* ptr = nullptr;
    vn = strtol(opt[1].c_str(), &ptr, 0);
    if (vn < 1) err = badopt(vpname, optstr, USAGE, 1.0, 100.0);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "t")       // 時期項目の値指定
  {
    if (strsplit(tlst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(int ni)       // オプションをチェックする
{
  if (ni < 3) return disperr(vpname, "At least 3 items are required !");
  int p = ni - 1;
  if (p < vn + 1) return disperr(vpname, "Not enough items !");
  vtvec.resize(p);
  for (int j = 0; j < p; ++j) vtvec[j] = (double)j * vdt;
  if (settvec(vtvec, tlst, p) < 0) return disperr(vpname, "Wrong relative time of items !");
  if (set_res() < 0) return -2; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["atab"] = vresmap["cc"] = vresmap["statg"] = vresmap["statt"] = true;
  vresmap["ma"] = vresmap["md"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["atab"] = vresmap["cc"] = vresmap["statg"] = vresmap["statt"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["atab"] = vresmap["cc"] = vresmap["statg"] = vresmap["statt"] = true;
      }
      else if (res == "all")    // 全て出力
      {
        vresmap["def"] = vresmap["atab"] = vresmap["cc"] = vresmap["ma"] = vresmap["md"] = vresmap["statg"] = vresmap["statt"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (res == "m")      // -res=mと指定された時
      { vresmap["ma"] = vresmap["md"] = true; }
      else if (res[0] == 'm' && res[1] == 'a' && res.size() > 2)
      {
        switch (res[2])
        {
        case 'b' : vma = mcbon;
          break;
        case 'f' : vma = mcfis;
          break;
        case 's' : vma = mcsch;
          break;
        case 't' : vma = mctuk;
          break;
        default : return disperr(vpname, "Wrong -res=" + res + " option !");
        }
        vresmap["ma"] = true;
      }
      else if (res[0] == 'm' && res[1] == 'd' && res.size() > 2)
      {
        switch (res[2])
        {
        case 'b' : vmd = mcbon;
          break;
        case 'f' : vmd = mcfis;
          break;
        case 's' : vmd = mcsch;
          break;
        default : return disperr(vpname, "Wrong -res=" + res + " option !");
        }
        vresmap["md"] = true;
      }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class gstat                     // 群別統計量クラス
{
private:
  vector<statv<double>> vystatvec; // yの基礎統計量ベクトル
  int vn;                       // 時期変数の次数
  int vpp;                      // データが存在する時期数
  int vnp;                      // 回帰用全例数
  mat<double> smat;             // 回帰計算用積和行列
  mat<double> pmat;             // 回帰計算用後の逆行列
  vecdbl vb;                    // 回帰係数ベクトル
  double vsb;                   // 回帰平方和
  double vsp;                   // 時期変動平方和(回帰用Syy)
public:
  gstat() : vystatvec(), vn(0), vpp(-1), vnp(0), smat(), pmat(), vb(), vsb(0.0), vsp(0.0)
  // デフォールトコンストラクタ
  {}
  gstat(size_t p0, int n0) : vystatvec(p0 + 1), vn(n0), vpp(-1), vnp(0), smat(n0 + 2), pmat(n0 + 2), vb(n0 + 1), vsb(0.0), vsp(0.0)
  // コンストラクタ
  {}
  gstat(const gstat& gstat0) : vystatvec(gstat0.vystatvec), vn(gstat0.vn), vpp(gstat0.vpp), vnp(gstat0.vnp), smat(gstat0.smat), pmat(gstat0.pmat), vb(gstat0.vb), vsb(gstat0.vsb), vsp(gstat0.vsp)
  // コピーコンストラクタ
  {}
  virtual ~gstat()              // デストラクタ
  {}
  const gstat& operator=(const gstat& gstat0); // コピー代入演算子
  const gstat& operator+=(const gstat& gstat0); // 加算代入演算子
  double sigma(const vecdbl& fdatavec);         // データを累積する
  int n() const                 // 時期変数の次数を返す
  { return vn; }
  const statv<double>& ystat(int idx) const // データの基礎統計量を返す
  { return vystatvec[idx]; }
  int ystatn() const            // yの全例数を返す
  { return vystatvec[vystatvec.size() - 1].n(); }
  double ystatvr() const        // yの全分散を返す
  { return vystatvec[vystatvec.size() - 1].vr(); }
  double ystatm() const         // yの総平均を返す
  { return vystatvec[vystatvec.size() - 1].mean(); }
  double ystatss() const        // yの全平方和を返す
  { return vystatvec[vystatvec.size() - 1].ss(); }
  int pp();                     // データのある時期数を返す
  void setsmat(const option& opt); // 回帰式用行列をセットする
  void calsmat(bool calm);         // 回帰用平均値を計算する
  void calreg(bool calr);          // 回帰式を計算する
  int np() const                   // 回帰用全例数を返す
  { return vnp; }
  double mt(int idx) const      // tの平均を返す
  { return (0 <= idx && idx <= vn) ? smat(0, idx) : 0.0; }
  double my() const             // 回帰計算用yの平均を返す
  { return smat(0, smat.c() - 1); }
  double b(int idx) const       // 偏回帰係数を返す
  { return (0 <= idx && idx <= vn) ? vb[idx] : 0.0; }
  double sb() const             // 回帰平方和を返す
  { return vsb; }
  double sp() const             // 時期変動平方和(回帰用Syy)を返す
  { return vsp; }
  double r2() const             // 回帰式の重寄与率を返す
  { return (vsp > 0.0) ? vsb / vsp : 0.0; }
  double vxr(const vec<double>& dxv) const; // 多重比較用dV^(-1)dを返す
};

const gstat& gstat::operator=(const gstat& gstat0) // コピー代入演算子
{
  if (&gstat0 == this) return *this; // 自分自身への代入を防ぐ
  vystatvec = gstat0.vystatvec;  vn = gstat0.vn;  vpp = gstat0.vpp;
  vnp = gstat0.vnp;  smat = gstat0.smat;  pmat = gstat0.pmat;
  vb = gstat0.vb;  vsb = gstat0.vsb;  vsp = gstat0.vsp;
  return *this;                 // 自分自身を返す
}

const gstat& gstat::operator+=(const gstat& gstat0) // 加算代入演算子
{
  for (size_t j = 0; j < vystatvec.size(); ++j) vystatvec[j] += gstat0.vystatvec[j];
  smat += gstat0.smat;  pmat += gstat0.pmat;
  return *this;                 // 自分自身を返す
}

double gstat::sigma(const vecdbl& fdatavec) // データを累積する
{
  int n = 0;
  double sum = 0.0;
  int p = vystatvec.size() - 1;
  for (int j = 0; j < p; ++j)
  {
    if (fdatavec[j] != BLANKf)
    { vystatvec[j].sigma(fdatavec[j]);  ++n;  sum += fdatavec[j]; }
  }
  return (n > 0) ? sum * sum / (double)n : 0.0; // 個体平方和用の値を返す
}

int gstat::pp()                 // データのある時期数を返す
{
  if (vpp < 0)                  // vpp未集計
  {
    vpp = 0;
    int p = vystatvec.size() - 1;
    for (int j = 0; j < p; ++j)
    {
      if (vystatvec[j].n() > 0) vpp++;
    }
  }
  return vpp;
}

void gstat::setsmat(const option& opt) // 回帰式用行列をセットする
{
  int d = vn + 1;
  int p = vystatvec.size() - 1;
  double r = (double)((opt.am() == am0) ? 1 : vystatvec[0].n());
  smat = 0.0;
  vecdbl tyvec(d + 1);
  tyvec[0] = 1.0;
  for (int j = 0; j < p; ++j)
  {
    if (vystatvec[j].n() > 0)
    {
      double mj = vystatvec[j].mean(), tj = opt.t(j), tt = opt.t(j);
      vystatvec[p] += vystatvec[j];  tyvec[d] = mj;
      for (int k = 0; k < vn; ++k)
      {
        tyvec[k + 1] = tt;
        tt *= tj;
      }
      if (opt.am() == am1 || opt.am() == amnon) r = vystatvec[j].n();
      for (int i = 0; i <= d; ++i)
      {
        double t = r * tyvec[i];
        for (int k = i; k <= d; ++k) smat.add(i, k, t * tyvec[k]);
      }
    }
  }
  pmat = smat;
}

void gstat::calsmat(bool calm)  // 回帰用平均値を計算する
{
  for (size_t i = 0; i < smat.r() - 1; ++i)
  {
    for (size_t j = i + 1; j < smat.c(); ++j)
    { smat.assign(j, i, smat(i, j));  pmat.assign(j, i, pmat(i, j)); }
  }
  if (!calm) return;            // smatを計算しない(共通回帰用)
  vnp = (int)(smat(0, 0));
  smat.sweep(0);
  pmat = smat;
  vsp = smat(smat.r() - 1, smat.c() - 1);
  const double Veps = DBL_EPSILON * fmax(5.0, smat.maxabse());
  if (vsp < Veps) vsp = 0.0;
}

void gstat::calreg(bool calr)   // 回帰式を計算する
{
  int d = vn + 1;
  vsb = pmat(d, d);
  for (int i = 1; i < d; ++i) pmat.sweep(i);
  vsb -= pmat(d, d);
  const double Veps = DBL_EPSILON * fmax(5.0, pmat.maxabse());
  if (vsb < Veps) vsb = 0.0;
  if (!calr) return;            // 回帰係数を保存しない
  for (int i = 0; i < d; ++i) vb[i] = pmat(i, d);
}

double gstat::vxr(const vec<double>& dxv) const // 多重比較用dV^(-1)dを返す
{
  int d = dxv.dim();
  vec<double> tv = pmat.sub(1, 1, d, d) * dxv;
  return inp(dxv, tv);
}

using mapsg = map<string, gstat>;         // 群別統計量マップの別名定義
using mapsg_itr = mapsg::iterator;        // 反復子の別名定義
using mapsg_citr = mapsg::const_iterator; // const反復子の別名定義

class varclass                  // 変数クラス
{
private:
  int vnp;                      // 時期項目数
  mapsg vgsmap;                 // 群統計量マップ
  double vss;                   // 個体間平方和
  int vdfs;                     // 個体間自由度
  double vssr;                  // 個体残差平方和
  int vdfsr;                    // 個体残差自由度
  double vsa;                   // 群差平方和
  double vsaa;                  // 修正群差平方和
  int vdfa;                     // 群差自由度
  double vsbc;                  // 共通回帰平方和
  double vsb0;                  // 全体回帰平方和
  int vdfb;                     // 回帰自由度
  double vsd;                   // 非平行性平方和
  int vdfd;                     // 非平行性自由度
  double vsp;                   // 時期平方和
  int vdfp;                     // 時期自由度
  double vsl;                   // ズレ平方和
  int vdfl;                     // ズレ自由度
  double vsr;                   // 残差平方和
  int vdfr;                     // 残差自由度
  double vst;                   // 全体平方和
  int vdft;                     // 全体自由度
  gstat vgstatall;              // 全体の統計量オブジェクト
  gstat vgstatcom;              // 共通の統計量オブジェクト
  void cal_tcova(double& sa, int& dfa, double &ssr, int& dfsr, double& ss, int& dfs, double& sbc, double& sb0, int& dfb, double& saa, double& sd, int& dfd, double& sp, int& dfp, double& sl, int& dfl, double& sr, int& dfr, double& st, int& dft, const option& opt);
  // 共分散分析用各種パラメターを計算する
public:
  varclass(size_t ni, int n0) : vnp(ni - 1), vgsmap(), vss(0.0), vdfs(0), vssr(0.0), vdfsr(0), vsa(0.0), vsaa(0.0), vdfa(0), vsbc(0.0), vsb0(0.0), vdfb(0), vsd(0.0), vdfd(0), vsl(0.0), vdfl(0), vsr(0.0), vdfr(0), vst(0.0), vdft(0), vgstatall(ni - 1, n0), vgstatcom(ni - 1, n0) // コンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int np() const                // 時期項目数を返す
  { return vnp; }
  const mapsg& gsmap() const    // 群統計量マップを返す
  { return vgsmap; }
  int ng() const                // 群数を返す
  { return (int)(vgsmap.size()); }
  const gstat& gstatall() const // 全体の統計量オブジェクトを返す
  { return vgstatall; }
  const gstat& gstatcom() const // 共通の統計量オブジェクトを返す
  { return vgstatcom; }
  double ss() const             // 個体間平方和を返す
  { return vss; }
  int dfs() const               // 個体間自由度を返す
  { return vdfs; }
  double ssr() const            // 個体残差平方和を返す
  { return vssr; }
  int dfsr() const              // 個体残差自由度を返す
  { return vdfsr; }
  double sa() const             // 群差平方和を返す
  { return vsa; }
  double saa() const            // 修正群差平方和を返す
  { return vsaa; }
  int dfa() const               // 群差自由度を返す
  { return vdfa; }
  double sbc() const            // 共通回帰平方和を返す
  { return vsbc; }
  double sb0() const            // 全体回帰平方和を返す
  { return vsb0; }
  int dfb() const               // 回帰自由度を返す
  { return vdfb; }
  double sd() const             // 非平行性平方和を返す
  { return vsd; }
  int dfd() const               // 非平行性自由度を返す
  { return vdfd; }
  double sp() const             // 時期平方和を返す
  { return vsp; }
  int dfp() const               // 時期自由度を返す
  { return vdfp; }
  double sl() const             // ズレ平方和を返す
  { return vsl; }
  int dfl() const               // ズレ自由度を返す
  { return vdfl; }
  double sr() const             // 残差平方和を返す
  { return vsr; }
  int dfr() const               // 残差自由度を返す
  { return vdfr; }
  double st() const             // 全体平方和を返す
  { return vst; }
  int dft() const               // 全体自由度を返す
  { return vdft; }
  int cal(const option& opt, ldata& datalst); // データを集計して統計量を求める
};

void varclass::cal_tcova(double& sa, int& dfa, double &ssr, int& dfsr, double& ss, int& dfs, double& sbc, double& sb0, int& dfb, double& saa, double& sd, int& dfd, double& sp, int& dfp, double& sl, int& dfl, double& sr, int& dfr, double& st, int& dft, const option& opt)
// 共分散分析用各種パラメターを計算する
{
  int p = vnp, g = vgsmap.size();
  double ssp = 0.0, ssb = 0.0;
  dfp = 0;  sa = sl = 0.0;
  for (mapsg_itr gitr = vgsmap.begin(); gitr != vgsmap.end(); ++gitr)
  {
    dfp += gitr->second.pp();
    sa += gitr->second.np() * gitr->second.my() * gitr->second.my();
    ssp += gitr->second.sp();  ssb += gitr->second.sb();
    sl += gitr->second.sp() - gitr->second.sb();
  }
  const double Seps = DBL_EPSILON * fmax(5.0, ssp);
  if (sl < Seps) sl = 0.0;
  --dfp;  sb0 = vgstatall.sb();  sbc = vgstatcom.sb();
  dfb = opt.n();
  double cf = vgstatall.np() * vgstatall.my() * vgstatall.my();
  sa -= cf;  dfa = g - 1;  ss -= cf;  dfs = vgstatall.ystat(0).n() - 1;
  if (sa < Seps) sa = 0.0;
  if (ss < Seps) ss = 0.0;
  sp = sa + ssp;  saa = sa + sbc - sb0;  sd = ssb - sbc;
  if (saa < Seps) saa = 0.0;
  if (sd < Seps) sd = 0.0;
  dfd = dfb * dfa;  dfl = g * ((p - 1) - dfb);
  if (dfl < 1 || sl < Seps) sl = 0.0;
  switch (opt.am())
  {
  case am0 : st = vgstatall.sp();  dft = vgstatall.np() - 1;
    ssr = sr = st - sa - ssb;  dfsr = dfr = dft - dfa - g * dfb;
    break;
  case am1 : st = vgstatall.ystatss();  dft = vgstatall.ystatn() - 1;
    ssr = sr = st - sp;  dfsr = dfr = dft - dfp;
    break;
  default : st = vgstatall.ystatss();  dft = vgstatall.ystatn() - 1;
    ssr = ss - sa;  dfsr = dfs - dfa;
    sr = st - ss - ssp;  dfr = dft - dfs - g * (p - 1);
  }
  if (ssr < Seps) ssr = 0.0;
  if (dfr < 1 || sr < Seps) sr = 0.0;
}

int varclass::cal(const option& opt, ldata& datalst)
// データを集計して統計量を求める
{
  vecdbl fdatavec(vnp);
  vss = 0.0;
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    bool blank = false;
    for (int j = 0; j < vnp; ++j)
    {
      fdatavec[j] = ditr->fdata(j + 1);
      if (fdatavec[j] == BLANKf) blank = true;
    }
    if (ditr->sdata(0) != BLANKs && (blank == false || opt.am() != am2))
    {
      mapsg_itr gitr = vgsmap.find(ditr->sdata(0));
      if (gitr == vgsmap.end()) // 新規群コード
      {
        gstat gstati(vnp, opt.n());
        vss += gstati.sigma(fdatavec);
        vgsmap.insert(make_pair(ditr->sdata(0), gstati));
      }
      else                      // すでにある群
      { vss += gitr->second.sigma(fdatavec); }
    }
  }
  if (vgsmap.empty()) return 0; // データなし
  for (mapsg_itr gitr = vgsmap.begin(); gitr != vgsmap.end(); ++gitr)
  {
    gitr->second.setsmat(opt);
    vgstatall += gitr->second;
    gitr->second.calsmat(true);
    vgstatcom += gitr->second;
    gitr->second.calreg(true);
  }
  vgstatall.calsmat(true);  vgstatcom.calsmat(false);
  vgstatall.calreg(true);  vgstatcom.calreg(true);
  cal_tcova(vsa, vdfa, vssr, vdfsr, vss, vdfs, vsbc, vsb0, vdfb, vsaa, vsd, vdfd, vsp, vdfp, vsl, vdfl, vsr, vdfr, vst, vdft, opt);//gstatall, gstatcom);
  return (int)(vgsmap.size());  // 群数を返す
}

class tcovaclass              // tcovaメインクラス
{
private:
  string fname;                 // ファイル名
  void put_id(ostream& fout, idd& id) const;
  // タイトルとデータ識別情報を出力する
  void put_mreg(ostream& fout, option& opt, const idd& id, const varclass& var) const;
  // 群別基礎統計量と回帰式を出力する
  void put_mregtab(ostream& fout, const option& opt, const idd& id, const gstat& gstati, const gstat& gstatcom) const;
  // 基礎統計量と回帰式の表を出力する
  void put_tcovatab0(ostream& fout, const varclass& var) const;
  // 単純な共分散分析表を出力する
  void put_tcovatab1(ostream& fout, const varclass& var) const;
  // 一元配置型共分散分析表を出力する
  void put_tcovatab1rn(ostream& fout, double sa, int dfa, double sbc, double sb0, int dfb, double saa, double sd, int dfd, double sl, int dfl, double sp, int dfp, double sr, int dfr, double st, int dft) const;
  // 繰り返しの有る／無い一元配置型共分散分析表を出力する
  void put_tcovatab2(ostream& fout, const varclass& var) const;
  // 二元配置型共分散分析表を出力する
  void put_tcovatabrm(ostream& fout, double sa, int dfa, double ssr, int dfsr, double ss, int dfs, double sb, int dfb, double sd, int dfd, double sl, int dfl, double sr, int dfr, double st, int dft) const;
  // 繰り返し測定二元配置型共分散分析表を出力する
  void put_tcovatab2f(ostream& fout, double sa, int dfa, double ssr, int dfsr, double ss, int dfs, double sb, int dfb, double sd, int dfd, double sl, int dfl, double sr, int dfr, double st, int dft) const;
  // 繰り返しの無い二元配置型共分散分析表を出力する
  void put_macmp(ostream& fout, const option& opt, const varclass& var) const;
  // 修正群差の多重比較を出力する
  void put_maci(ostream& fout, const option& opt, const varclass& var) const;
  // 修正群差の信頼区間を出力する
  void put_mdcmp(ostream& fout, const option& opt, const idd& id, const varclass& var) const;
  // 非平行性の多重比較を出力する
public:
  tcovaclass() : fname()        // コンストラクタ
  {}
  ~tcovaclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込み保存・累積する
  void put(option& opt, idd& id, ldata& datalst) const;
  // 集計結果を出力する
};

void tcovaclass::put_id(ostream& fout, idd& id) const
// タイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "GROUP ITEM   :" << id.item(0) << "\n";
  for (size_t i = 1; i < id.ni(); ++i)
  {
    fout << "PERIOD ITEM" << setw(2) << i;
    fout << ":" << id.item(i) << "\n";
  }
}

void tcovaclass::put_mreg(ostream& fout, option& opt, const idd& id, const varclass& var) const
// 群別基礎統計量と回帰式を出力する
{
  if (opt.res("statg"))
  {
    int gno = 1;
    for (mapsg_citr gitr = (var.gsmap()).begin(); gitr != (var.gsmap()).end(); ++gitr, ++gno)
    {
      fout << "\nGROUP" << setw(4) << gno << ":" << id.item(0) << "=" << gitr->first << "\n";
      put_mregtab(fout, opt, id, gitr->second, var.gstatcom());
    }
  }
  if (opt.res("statt"))
  {
    fout << "\n<OVERALL>\n";  // 全体はvspが不適切な値になるので出力しない
    put_mregtab(fout, opt, id, var.gstatall(), var.gstatcom());
  }
}

void tcovaclass::put_mregtab(ostream& fout, const option& opt, const idd& id, const gstat& gstati, const gstat& gstatcom) const
// 基礎統計量と回帰式の表を出力する
{
  constexpr int LENline = 78;   // 区切り線の長さ
  constexpr int Nterm = 4;      // 1行あたりの項数
  int p = id.ni() - 1;
  putline(fout, LENline);
  for (int j = 0; j < p; ++j)
  {
    fout << "t=" << setw(7) << opt.t(j);
    fout << ":";  putmsd(fout, gstati.ystat(j).n(), gstati.ystat(j).mean(), gstati.ystat(j).vr());
  }
  putline(fout, LENline);
  fout << "ALL      :";
  putmsd(fout, gstati.ystatn(), gstati.ystatm(), gstati.ystatvr());
  double bc0 = gstati.my();
  fout << "Group Regression Equation :y=" << setw(13) <<  gstati.b(0);
  for (int j = 1; j <= opt.n(); ++j)
  {
    if ((j + 1) % Nterm == 1) fout << "\n" << setw(13) << "";
    fout << showpos << setw(9) << gstati.b(j) << noshowpos;
    fout << "t";
    if (opt.n() > 1) fout << "^" << left << setw(2) << j << right;
    bc0 -= gstatcom.b(j) * gstati.mt(j);
  }
  fout << "\nCommon Regression Equation:y=" << setw(13) << bc0;
  for (int j = 1; j <= opt.n(); ++j)
  {
    if ((j + 1) % Nterm == 1) fout << "\n" << setw(13) << "";
    fout << showpos << setw(9) << gstatcom.b(j) << noshowpos;
    fout << "t";
    if (opt.n() > 1) fout << "^" << left << setw(2) << j << right;
  }
  fout << "\nContribution Ratio of Regression Equation:r-squared=" << gstati.r2() << endl;
}

void tcovaclass::put_tcovatab0(ostream& fout, const varclass& var) const
// 単純な共分散分析表を出力する
{
  constexpr int LENline = 78;   // 区切り線の長さ
  string tstr("ANCOVA Table");
  fout << "\n" << setw((LENline - strclen(tstr, CHARwide)) / 2 + tstr.size()) << tstr;
  fout << "\nFACTOR    Sum of Squares        DF  Mean Squares     F-value       p-value\n";
  putline(fout, LENline);
  double v = 0.0 , f = 0.0 , pr = 0.0;
  double vr = calvfp(v, f, pr, var.sa(), var.dfa(), var.sr(), var.dfr());
  tstr = "GROUP";
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.sa();
  fout << setw(10) << var.dfa() << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  calvfp(v, f, pr, var.sbc(), var.dfb(), var.sr(), var.dfr());
  tstr = "COM-REGG";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.sbc();
  fout << setw(10) << var.dfb() << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, var.saa(), var.dfa(), var.sr(), var.dfr());
  tstr = "ADJ-GROUP";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.saa();
  fout << setw(10) << var.dfa() << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  calvfp(v, f, pr, var.sb0(), var.dfb(), var.sr(), var.dfr());
  tstr = "ALL-REG";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.sb0();
  fout << setw(10) << var.dfb() << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, var.sd(), var.dfd(), var.sr(), var.dfr());
  tstr = "NONPARALLE";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.sd();
  fout << setw(10) << var.dfd() << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  tstr = "RESIDUALS";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.sr();
  fout << setw(10) << var.dfr() << setw(14) << vr;
  fout << "\n";
  putline(fout, LENline);
  tstr = "TOTAL";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.st();
  fout << setw(10) << var.dft();
  fout << "\n";
}

void tcovaclass::put_tcovatab1(ostream& fout, const varclass& var) const
// 一元配置型共分散分析表を出力する
{
  if (var.dfr() > 0)            // 繰り返しの有る一元配置型
  {
    put_tcovatab1rn(fout, var.sa(), var.dfa(), var.sbc(), var.sb0(), var.dfb(), var.saa(), var.sd(), var.dfd(), var.sl(), var.dfl(), var.sp(), var.dfp(), var.sr(), var.dfr(), var.st(), var.dft());
  }
  else                          // 繰り返しの無い一元配置型
  {
    double sl = 0.0, st = var.sp();
    int dfl = 0, dft = var.dfp();
    double sr = var.st() - var.sa() - var.sbc() - var.sd();
    int dfr = var.dft() - var.dfa() - var.dfb() - var.dfd();
    put_tcovatab1rn(fout, var.sa(), var.dfa(), var.sbc(), var.sb0(), var.dfb(), var.saa(), var.sd(), var.dfd(), sl, dfl, var.sp(), var.dfp(), sr, dfr, st, dft);
  }
}

void tcovaclass::put_tcovatab1rn(ostream& fout, double sa, int dfa, double sbc, double sb0, int dfb, double saa, double sd, int dfd, double sl, int dfl, double sp, int dfp, double sr, int dfr, double st, int dft) const
// 繰り返しの有る／無い一元配置型共分散分析表を出力する
{
  constexpr int LENline = 78;
  string tstr("ANOVA1+ANCOVA Table");
  fout << "\n" << setw((LENline - strclen(tstr, CHARwide)) / 2 + tstr.size()) << tstr;
  fout << "\nFACTOR    Sum of Squares        DF  Mean Squares     F-value       p-value\n";
  putline(fout, LENline);
  double v = 0.0, f = 0.0, pr = 0.0;
  double vr = calvfp(v, f, pr, sa, dfa, sr, dfr);
  tstr = "GROUP";
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sa;
  fout << setw(10) << dfa << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  calvfp(v, f, pr, sbc, dfb, sr, dfr);
  tstr = "COMMON-REG";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sbc;
  fout << setw(10) << dfb << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, saa, dfa, sr, dfr);
  tstr = "ADJ-GROUP";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << saa;
  fout << setw(10) << dfa << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  calvfp(v, f, pr, sb0, dfb, sr, dfr);
  tstr = "ALL-REG";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sb0;
  fout << setw(10) << dfb << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, sd, dfd, sr, dfr);
  tstr = "NONPARALLE";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sd;
  fout << setw(10) << dfd << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  if (dfl > 0)
  {
    calvfp(v, f, pr, sl, dfl, sr, dfr);
    tstr = "TOTAL LOF";
    npad = tstr.size() - strclen(tstr, CHARwide);
    fout << left << setw(10 + npad) << tstr << right << setw(14) << sl;
    fout << setw(10) << dfl << setw(14) << v << setw(12) << f << setw(14) << pr;
    fout << pind(pr) << "\n";
    putline(fout, LENline);
    calvfp(v, f, pr, sp, dfp, sr, dfr);
    tstr = "PERIOD";
    npad = tstr.size() - strclen(tstr, CHARwide);
    fout << left << setw(10 + npad) << tstr << right << setw(14) << sp;
    fout << setw(10) << dfp << setw(14) << v << setw(12) << f << setw(14) << pr;
    fout << pind(pr) << "\n";
  }
  tstr = "RESIDUALS";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sr;
  fout << setw(10) << dfr << setw(14) << vr;
  fout << "\n";
  putline(fout, LENline);
  tstr = "TOTAL";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << st;
  fout << setw(10) << dft;
  fout << endl;
}

void tcovaclass::put_tcovatab2(ostream& fout, const varclass& var) const
// 二元配置型共分散分析表を出力する
{
  if (var.dfsr() > 0)           // 繰り返し測定二元配置型
  {
    put_tcovatabrm(fout, var.sa(), var.dfa(), var.ssr(), var.dfsr(), var.ss(), var.dfs(), var.sb0(), var.dfb(), var.sd(), var.dfd(), var.sl(), var.dfl(), var.sr(), var.dfr(), var.st(), var.dft());
  }
  else                          // 繰り返しの無い二元配置型
  {
    put_tcovatab2f(fout, var.sa(), var.dfa(), var.ssr(), var.dfsr(), var.ss(), var.dfs(), var.sb0(), var.dfb(), var.sd(), var.dfd(), var.sl(), var.dfl(), var.sr(), var.dfr(), var.st(), var.dft());
  }
}

void tcovaclass::put_tcovatabrm(ostream& fout, double sa, int dfa, double ssr, int dfsr, double ss, int dfs, double sb, int dfb, double sd, int dfd, double sl, int dfl, double sr, int dfr, double st, int dft) const
// 繰り返し測定二元配置型共分散分析表を出力する
{
  constexpr int LENline = 78;   // 区切り線の長さ
  string tstr("RMANOVA+ANCOVA Table");
  fout << "\n" << setw((LENline - strclen(tstr, CHARwide)) / 2 + tstr.size()) << tstr;
  fout << "\nFACTOR    Sum of Squares        DF  Mean Squares     F-value       p-value\n";
   putline(fout, LENline);
  double v = 0.0 , f = 0.0 , pr = 0.0;
  double vr = calvfp(v, f, pr, sa, dfa, ssr, dfsr);
  tstr = "GROUP";
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sa;
  fout << setw(10) << dfa << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  tstr = "SUBJ-RES";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << ssr;
  fout << setw(10) << dfsr << setw(14) << vr;
  fout << "\n";
  putline(fout, LENline);
  vr = calvfp(v, f, pr, ss, dfs, sr, dfr);
  tstr = "SUBJECT";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << ss;
  fout << setw(10) << dfs << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  calvfp(v, f, pr, sb, dfb, sr, dfr);
  tstr = "ALL-REG";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sb;
  fout << setw(10) << dfb << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  calvfp(v, f, pr, sd, dfd, sr, dfr);
  tstr = "NONPARALLE";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sd;
  fout << setw(10) << dfd << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  calvfp(v, f, pr, sl, dfl, sr, dfr);
  tstr = "TOTAL LOF";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sl;
  fout << setw(10) << dfl << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  tstr = "RESIDUALS";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sr;
  fout << setw(10) << dfr << setw(14) << vr;
  fout << "\n";
  putline(fout, LENline);
  tstr = "TOTAL";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right;
  fout << setw(14) << st;
  fout << setw(10) << dft;
  fout << endl;
}

void tcovaclass::put_tcovatab2f(ostream& fout, double sa, int dfa, double ssr, int dfsr, double ss, int dfs, double sb, int dfb, double sd, int dfd, double sl, int dfl, double sr, int dfr, double st, int dft) const
// 繰り返しの無い二元配置型共分散分析表を出力する
{
  constexpr int LENline = 78;   // 区切り線の長さ
  string tstr("ANOVA2+ANCOVA Table");
  fout << "\n" << setw((LENline - strclen(tstr, CHARwide)) / 2 + tstr.size()) << tstr;
  fout << "\nFACTOR    Sum of Squares        DF  Mean Squares     F-value       p-value\n";
  putline(fout, LENline);
  double v = 0.0 , f = 0.0 , pr = 0.0;
  double vr = calvfp(v, f, pr, ss, dfs, sl, dfl);
  tstr = "GROUP(SUB)";
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << ss;
  fout << setw(10) << dfs << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  calvfp(v, f, pr, sb, dfb, sl, dfl);
  tstr = "ALL-REG";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sb;
  fout << setw(10) << dfb << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  calvfp(v, f, pr, sd, dfd, sl, dfl);
  tstr = "NONPARALLE";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sd;
  fout << setw(10) << dfd << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  tstr = "RESIDUALS";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sl;
  fout << setw(10) << dfl << setw(14) << vr;
  fout << "\n";
  putline(fout, LENline);
  tstr = "TOTAL";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << st;
  fout << setw(10) << dft;
  fout << "\n";
}

void tcovaclass::put_macmp(ostream& fout, const option& opt, const varclass& var) const
// 修正群差の多重比較を出力する
{
  if (var.ssr() <= 0.0) return; // 出力なし
  int g = var.ng(), df1 = 1, dc = 1;
  string method = "TUKEY";
  switch (opt.ma())
  {
  case mcbon : method = "BONFERRONI";  dc = g * (g - 1) / 2;
    break;
  case mcfis : method = "FISHER";
    break;
  case mcsch : method = "SCHEFFE";  df1 = g - 1;
  }
  fout << "\n<" << method << " TYPE MULTIPLE COMPARISON>\n";
  fout << "Covariate=Overall Mean  ";
  if (opt.ma() == mctuk)
  {
    fout << "GROUPS=" << g << "  DF=" << var.dfsr() << "\n";
    fout << " G:    Adj-Mean -  G:    Adj-Mean    Adj-Diff     q-value       p-value\n";
  }
  else
  {
    fout << "DF1=" << df1 << "  DF2=" << var.dfsr() << "\n";
    fout << " G:    Adj-Mean -  G:    Adj-Mean    Adj-Diff     F-value       p-value\n";
  }
  constexpr int LENline = 75;   // 区切り線の長さ
  putline(fout, LENline);
  double vr = var.ssr() / var.dfsr();
  vec<double> dxv(opt.n());
  double yc = 0.0;
  for (int k = 1; k <= opt.n(); ++k) yc += (var.gstatcom()).b(k) * (var.gstatall()).mt(k);
  mapsg_citr gitri = (var.gsmap()).begin();
  for (int i = 0; i < g - 1; ++i, ++gitri)
  {
    const gstat* gstatpi = &(gitri->second);
    double ni = (double)(gstatpi->np());
    double bc0i = gstatpi->my();
    for (int k = 1; k <= opt.n(); ++k) bc0i -= (var.gstatcom()).b(k) * gstatpi->mt(k);
    double yi = bc0i + yc;
    mapsg_citr gitrj = gitri;
    ++gitrj;
    int j = i + 1; 
    while (gitrj != (var.gsmap()).end())
    {
      const gstat* gstatpj = &(gitrj->second);
      double bc0j = gstatpj->my();
      for (int k = 1; k <= opt.n(); ++k)
      {
        bc0j -= (var.gstatcom()).b(k) * gstatpj->mt(k);
        dxv.assign(k - 1, gstatpj->mt(k) - gstatpi->mt(k));
      }
      double yj = bc0j + yc;
      double dy = bc0j - bc0i;
      double vd = (1.0 / ni + 1.0 / (double)(gstatpj->np()) + (var.gstatall()).vxr(dxv)) * vr;
      double f = dy * dy / vd / (double)df1;
      double pr = calmcpr(opt.ma(), f, g, df1, var.dfsr(), dc);
      fout << setw(2) << j + 1;
      fout << ":";
      fout << setw(12) << yj;
      fout << " -";
      fout << setw(3) << i + 1;
      fout << ":";
      fout << setw(12) << yi << setw(12) << dy << setw(12) << f << setw(14) << pr;
      fout << pind(pr) << "\n";
      ++gitrj;  ++j;
    }
  }
  putline(fout, LENline);
}

void tcovaclass::put_maci(ostream& fout, const option& opt, const varclass& var) const
// 修正群差の信頼区間を出力する
{
  if (var.ssr() <= 0.0) return; // 出力なし
  int g = var.ng(), df1 = 1, dc = 1;
  string method = "TUKEY";
  switch (opt.ma())
  {
  case mcbon : method = "BONFERRONI";  dc = g * (g - 1) / 2;
    break;
  case mcfis : method = "FISHER";
    break;
  case mcsch : method = "SCHEFFE";  df1 = g - 1;
  }
  fout << "\n<" << method << " TYPE " << opt.cc() * 100.0 << "% SIMULTANEOUS CONFIDENCE INTERVAL>\n";
  fout << " G -  G    Adj-Mean   Precision   Lower-Lim   Upper-Lim\n";
  constexpr int LENline = 55;   // 区切り線の長さ
  putline(fout, LENline);
  double vr = var.ssr() / var.dfsr();
  vec<double> dxv(opt.n());
  mapsg_citr gitri = (var.gsmap()).begin();
  for (int i = 0; i < g - 1; ++i, ++gitri)
  {
    const gstat* gstatpi = &(gitri->second);
    double ni = (double)(gstatpi->np());
    double bc0i = gstatpi->my();
    for (int k = 1; k <= opt.n(); ++k) bc0i -= (var.gstatcom()).b(k) * gstatpi->mt(k);
    mapsg_citr gitrj = gitri;
    ++gitrj;
    int j = i + 1;
    while (gitrj != (var.gsmap()).end())
    {
      const gstat* gstatpj = &(gitrj->second);
      double bc0j = gstatpj->my();
      for (int k = 1; k <= opt.n(); ++k)
      {
        bc0j -= (var.gstatcom()).b(k) * gstatpj->mt(k);
        dxv.assign(k - 1, gstatpi->mt(k) - gstatpj->mt(k));
      }
      double dy = bc0j - bc0i;
      double vd = (1.0 / ni + 1.0 / (double)(gstatpj->np()) + (var.gstatall()).vxr(dxv)) * vr;
      double ci = calmcci(opt.ma(), opt.cc(), g, df1, var.dfsr(), dc, vd);
      fout << setw(2) << j + 1;
      fout << " -";
      fout << setw(3) << i + 1;
      fout << setw(12) << dy;
      fout << setw(12) << ci << setw(12) << dy - ci << setw(12) << dy + ci;
      fout << "\n";
      ++gitrj;  ++j;
    }
  }
  putline(fout, LENline);
}

void tcovaclass::put_mdcmp(ostream& fout, const option& opt, const idd& id, const varclass& var) const
// 非平行性の多重比較を出力する
{
  if (var.sr() <= 0.0) return;  // 出力なし
  int g = var.ng(), p = id.ni() - 1;
  int df1 = opt.n(), dc = g * (g - 1) / 2;
  string method = "BONFERRONI";
  switch (opt.md())
  {
  case mcfis : method = "FISHER";  dc = 1;
    break;
  case mcsch : method = "SCHEFFE";  df1 = (g - 1) * opt.n();  dc = 1;
  }
  fout << "\n" << method << " TYPE MULTIPLE COMPARISON>\n";
  fout << "DF1=" << df1 << "  DF2=" << var.dfr() << "\n";
  fout << " G -  G   SS of Reg-Coeff     F-value       p-value\n";
  constexpr int LENline = 55;   // 区切り線の長さ
  putline(fout, LENline);
  double vr = var.sr() / var.dfr();
  vec<double> dbv(opt.n());
  mapsg_citr gitri = (var.gsmap()).begin();
  for (int i = 0; i < g - 1; ++i, ++gitri)
  {
    mapsg_citr gitrj = gitri;
    ++gitrj;
    int j = i + 1;
    while (gitrj != (var.gsmap()).end())
    {
      gstat gstatc(p, opt.n());
      gstatc += gitri->second;  gstatc += gitrj->second;
      gstatc.calreg(false);
      for (int k = 1; k <= opt.n(); ++k)
      { dbv.assign(k - 1, gitrj->second.b(k) - gitri->second.b(k)); }
      double sdb = gstatc.vxr(dbv);
      double f = sdb / vr / df1;
      double pr = calmcpr(opt.md(), f, g, df1, var.dfr(), dc);
      fout << setw(2) << j + 1;
      fout << " -";
      fout << setw(3) << i + 1;
      fout << setw(18) << sdb << setw(12) << f << setw(14) << pr;
      fout << pind(pr) << "\n";
      ++gitrj;  ++j;
    }
  }
  putline(fout, LENline);
}

int tcovaclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void tcovaclass::put(option& opt, idd& id, ldata& datalst) const
// 集計結果を出力する
{
  if (opt.res("res") == false) return;
  put_id(cout, id);
  if (id.nc() < 1)              // データなし
  {
    cout << "\n" << endl;
    return;
  }
  varclass var(id.ni(), opt.n());        // 変数オブジェクト
  if (var.cal(opt, datalst) < 1) return; // 群なし
  if (opt.res("statg") || opt.res("statt")) put_mreg(cout, opt, id, var);
  if (opt.res("atab"))
  {
    switch (opt.am())
    {
    case amnon :
      break;
    case am0 :
      put_tcovatab0(cout, var);
      break;
    case am1 :
      put_tcovatab1(cout, var);
      break;
    default :
      put_tcovatab2(cout, var);
    }
  }
  if (opt.res("ma"))
  {
    put_macmp(cout, opt, var);
    put_maci(cout, opt, var);
  }
  if (opt.res("md")) put_mdcmp(cout, opt, id, var);
  cout << "\n" << endl;
}

int main(int argc, char *argv[]) // メインルーチン
{
  tcovaclass tcova;             // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) tcova.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { tcova.get(argv[i], opt, id, datalst);  fok = false; }
    }
    tcova.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
