// trege.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-11 13:29:44 tac>
//
// 時系列回帰分析を行う。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -W trege trege.C
//               -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "trege  Version 7.41e         Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:treg  [<OPTION> ...]  [<DATA FILE> ...]",
  "  The time series regression analysis with all items.",
  "<OPTION>",
  "  -am={0|1|2}:select the method of ANOVA(default:-am=2)",
  "    0:simple method using only means  1:one-way layout method",
  "    2:two-way layout method(default)",
  "  -dt=<INTERVAL>:specify time interval(default:-dt=1)",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -ngp={mse|reg}[,...]:select the data for Ngraph",
  "    mse:mean +/- SE  reg:regresion line  default:all",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=trege_ngp.dat",
  "  -res=[-][{def|all|reg|stat}][,...]:select output results",
  "    -:wiht -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    reg:regression analysis  stat:basic statistics",
  "  -t=<TIME1>,<TIME2>[,<TIME3>,...]:specify the relative time of items"
};

const string Title = "=== Time Series Regression Analysis ==="; // 表題
const string NGPNAMEdef = "trege_ngp.dat"; // 省略時Ngraphファイル名

enum amtype                     // 周期解析の種類
{
  amnon,                        // なし
  am0,                          // 平均値を用いる単純法
  am1,                          // 一元配置法
  am2,                          // 二元配置法
  ametc                         // その他
};

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  amtype vam;                   // 分散分析の種類
  double vdt;                   // 時期間隔
  string vngpname;              // Ngraphファイル基本名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  lstr tlst;                    // 項目時間リスト
  vecdbl vtvec;                 // 項目時間ベクトル
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  amtype am() const             // 分散分析の種類を返す
  { return vam; }
  double dt() const             // 時期間隔を返す
  { return vdt; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 分析結果出力フラッグを返す
  double t(size_t idx) const    // 項目時間を返す
  { return (idx < vtvec.size()) ? vtvec[idx] : 0.0; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vdt(1.0), vngpname(), vngplst(), vngpmap(), vam(am2), vreslst(), vresmap(), tlst(), vtvec()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngpitr = vngpmap.find(code);
  if (ngpitr == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "am")           // 分散分析の手法指定
  {
    if (opt[1].empty() || opt[1] == "2") vam = am2;
    else if (opt[1] == "0") vam = am0;
    else if (opt[1] == "1") vam = am1;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "dt")      // 時期間隔の指定
  {
    vdt = strtodf(opt[1]);
    if (vdt <= 0.0 || vdt == BLANKf) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "t")       // 項目時間の指定
  {
    if (strsplit(tlst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのチェックとセット
{
  if (ni < 2) return disperr(vpname, "At least 2 items are required !");
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  vtvec.resize(ni);
  for (size_t j = 0; j < ni; ++j) vtvec[j] = (double)j * vdt;
  if (settvec(vtvec, tlst, ni) < 0) return disperr(vpname, "Wrong relative time of items !");
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["mse"] = vngpmap["reg"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["mse"] = vngpmap["reg"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngpitr = vngpmap.find(ngp);
      if (ngpitr == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["reg"] = vresmap["stat"] = true;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["reg"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def" || res == "all")
      {
        vresmap["def"] = vresmap["reg"] = vresmap["stat"] = true;
      }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class varclass                  // 変数クラス
{
private:
  int vnp;                      // 時期項目数
  vector<statv<double>> ystatvec; // 時期項目の基本統計量ベクトル
  int vnb;                        // 観察個体数
  double vsb;                     // 観察個体平方和
  statcor<double> statc;        // 相関基礎統計量
  double vb0;                   // 回帰式の定数
  double vb1;                   // 回帰係数
  double vst;                   // 全体平方和
  int vdft;                     // 全体自由度
  double vsa;                   // 群差平方和
  int vdfa;                     // 群差自由度
  int vdfb;                     // 回帰自由度
  double vsp;                   // 時期平方和
  int vdfp;                     // 時期自由度
  double vsl;                   // ズレ平方和
  int vdfl;                     // ズレ自由度
  double vsr;                   // 残差平方和
  int vdfr;                     // 残差自由度
  int cal_treg(const option& opt);
  // 基礎統計・時系列回帰式の各種パラメターを計算する
public:
  varclass(size_t ni) : vnp(ni), ystatvec(ni + 1), vnb(0), vsb(0.0), statc(), vb0(0.0), vb1(0.0), vst(0.0), vdft(0), vsa(0.0), vdfa(0), vdfb(0), vsp(0.0), vdfp(0), vsl(0.0), vdfl(0), vsr(0.0), vdfr(0) // コンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int np() const                // 時期項目数を返す
  { return vnp; }
  int nb() const                // 観察個体数を返す
  { return vnb; }
  int n(size_t idx) const       // 時期項目の例数を返す
  { return (idx < ystatvec.size()) ? ystatvec[idx].n() : 0; }
  double mean(size_t idx) const // 時期項目の平均を返す
  { return (idx < ystatvec.size()) ? ystatvec[idx].mean() : 0.0; }
  double vr(size_t idx) const   // 時期項目の不偏分散を返す
  { return (idx < ystatvec.size()) ? ystatvec[idx].vr() : 0.0; }
  double se(size_t idx) const   // 時期項目の標準誤差を返す
  { return (idx < ystatvec.size()) ? ystatvec[idx].se() : 0.0; }
  double rxy() const            // 相関係数を返す
  { return statc.rxy(); }
  double b0() const             // 回帰式の定数を返す
  { return vb0; }
  double b1() const             // 回帰係数を返す
  { return vb1; }
  double sb() const             // 観察個体平方和を返す
  { return vsb; }
  double st() const             // 全体平方和を返す
  { return vst; }
  int dft() const               // 全体自由度を返す
  { return vdft; }
  double sa() const             // 群差平方和を返す
  { return vsa; }
  int dfa() const               // 群差自由度を返す
  { return vdfa; }
  int dfb() const               // 回帰自由度を返す
  { return vdfb; }
  double sp() const             // 時期平方和を返す
  { return vsp; }
  int dfp() const               // 時期自由度を返す
  { return vdfp; }
  double sl() const             // ズレ平方和を返す
  { return vsl; }
  int dfl() const               // ズレ自由度を返す
  { return vdfl; }
  double sr() const             // 残差平方和を返す
  { return vsr; }
  int dfr() const               // 残差自由度を返す
  { return vdfr; }
  int cal(const option& opt, ldata& datalst); // データを集計して統計量を求める
};

int varclass::cal_treg(const option& opt)
// 基礎統計・時系列回帰式の各種パラメターを計算する
{
  int p = vnp, np = 0;
  vsa = 0.0;
  for (int j = 0; j < p; ++j)
  {
    int n = ystatvec[j].n();
    if (n > 0)
    {
      ystatvec[p] += ystatvec[j];
      double tj = opt.t(j), yj = ystatvec[j].mean();
      int r = (opt.am() == am0) ? 1 : n;
      for (int i = 0; i < r; ++i) statc.sigma(tj, yj);
      ++np;  vsa += n * yj * yj;
    }
  }
  if (np < 2) return disperr(opt.pname(), "too few periodic items !");
  double mx = statc.mx(), my = statc.my();
  double sxx = statc.sxx(), sxy = statc.sxy(), syy = statc.syy();
  vb1 = (sxx > 0.0) ? sxy / sxx : 0.0;
  vb0 = my - vb1 * mx;
  vst = (opt.am() == am0) ? syy : ystatvec[p].ss();
  vdfa = np - 1;
  vdft = (opt.am() == am0) ? vdfa : ystatvec[p].n() - 1;
  double cf = ystatvec[p].n() * ystatvec[p].mean() * ystatvec[p].mean();
  vsa -= cf;  vsb = vsb / (double)p - cf;  vsp = sxy * vb1;
  vsl = vsa - vsp;
  const double Seps = DBL_EPSILON * fmax(5.0, syy);
  if (vsa < Seps) vsa = 0.0;
  if (vsb < Seps) vsb = 0.0;
  if (vsp < Seps) vsp = 0.0;
  if (vsl < Seps) vsl = 0.0;
  vdfp = 1;
  vdfl = vdfa - vdfp;  vdfb = vnb - 1;
  if (vdfl < 1) vsl = 0.0;
  if (opt.am() == am1) { vsr = vst - vsa;  vdfr = vdft - vdfa; }
  else if (opt.am() == am2) { vsr = vst - vsa - vsb;  vdfr = vdft - vdfa - vdfb; }
  else { vsr = vst - vsp;  vdfr = vdft - vdfp; }
  if (vdfr < 1 || vsr < Seps) vsr = 0.0;
  return np;                    // 有効な時期項目数を返す
}

int varclass::cal(const option& opt, ldata& datalst)
// データを集計して統計量を求める
{
  vecdbl fdatavec(vnp);
  for (size_t j = 0; j < ystatvec.size(); ++j) ystatvec[j].clear();
  vnb = 0;  vsb = 0.0;
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    bool blank = false;
    for (int j = 0; j < vnp; ++j)
    {
      fdatavec[j] = ditr->fdata(j);
      if (fdatavec[j] == BLANKf) blank = true;
    }
    if (blank == false || opt.am() != am2)
    {
      double sum = 0.0;
      for (int j = 0; j < vnp; ++j)
      {
        if (fdatavec[j] != BLANKf)
        { ystatvec[j].sigma(fdatavec[j]);  sum += fdatavec[j]; }
      }
      ++vnb;  vsb += sum * sum;
    }
  }
  return cal_treg(opt);         // 有効な時期項目数を返す
}

class tregclass                 // tregメインクラス
{
private:
  option opt;                   // オプション
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  void put_resid(ostream& fout);   //解析結果のタイトルとデータ識別情報を出力する
  // 基礎統計・時系列回帰式の各種パラメターを計算する
  void put_stat(ostream& fout, const varclass& var) const;
  // 時期別基礎統計量を出力する
  void put_tregfun(ostream& fout, const varclass& var) const;
  // 時系列回帰式のパラメターを出力する
  void put_tregtab1(ostream& fout, double sp, int dfp, double sl, int dfl, double sa, int dfa, double sr, int dfr, double st, int dft) const;
  // 一元配置型分散分析表を出力する
  void put_tregtab2(ostream& fout, double sp, int dfp, double sl, int dfl, double sa, int dfa, double sb, int dfb, double sr, int dfr, double st, int dft) const;
  // 二元配置型分散分析表を出力する
  void plot_ngp(const varclass& var); // Ngraph用データを出力する
  void plot_ngpid(ofstream& fout);
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_ngpmse(ofstream& fout, const varclass& var) const;
  // Ngraph用時期別平均±SEを出力する
  void plot_ngpreg(ofstream& fout, const varclass& var) const;
  // Ngraph用時系列回帰直線を出力する
public:
  tregclass(const char* c_pname = nullptr) : opt(c_pname), fname(), id(), datalst()
  // コンストラクタ
  {}
  ~tregclass()                  // デストラクタ
  {}
  const string& pname() const   // プログラム名を返す
  { return opt.pname(); }
  void init()      // 初期設定を行う(コンストラクタ中ではできない処理)
  { opt.setdef(); }
  void setopt(const string& optstr, char devc = OPTdev) // オプションを設定する
  { opt.set(optstr, devc); }
  int get(const char* c_fname); // データを読み込み保存・累積する
  void put(int nc);             // 集計結果を出力する
};

void tregclass::put_resid(ostream& fout)
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "PERIOD ITEM  1:" << id.item(0) << "\n";
  fout << "  :\n";
  fout << "PERIOD ITEM" << setw(3) << id.ni();
  fout << ":" << id.item(id.ni() - 1) << "\n";
}

void tregclass::put_stat(ostream& fout, const varclass& var) const
// 時期別基礎統計量を出力する
{
  constexpr int LENline = 79;
  size_t p = id.ni();
  fout << "\nTIME t:BASIC STATISTICS\n";
  putline(fout, LENline);
  for (size_t j = 0; j < p; ++j)
  {
    int prec0 = fout.precision(2); // 小数点以下の精度を指定する
    fout.setf(ios::fixed);
    fout << setw(6) << opt.t(j) << setprecision(prec0); // 精度を元に戻す
    fout.unsetf(ios::fixed);
    fout << ":";
    putmsd(fout, var.n(j), var.mean(j), var.vr(j));
  }
  putline(fout, LENline);
  fout << "ALL   :";
  putmsd(fout, var.n(p), var.mean(p), var.vr(p));
}

void tregclass::put_tregfun(ostream& fout, const varclass& var) const
// 時系列回帰式のパラメターを出力する
{
  fout << "\nTime Series Regression Line:y=" << var.b0() << showpos << var.b1() << noshowpos << "t\n";
  double r2 = var.rxy() * var.rxy();
  if (opt.am() == am0)
  { fout << "Contribution Ratio of Regression Line:r-squared=" << r2 << endl; }
  else
  {
    r2 = (var.sa() > 0.0) ? var.sp() / var.sa() : 0.0;
    fout << "Contribution Ratio of Regression Line to Mean:r-squared=" << r2 << "\n";
    r2 = (var.st() > 0.0) ? var.sp() / var.st() : 0.0;
    fout << "Contribution Ratio of Regresion Line to Data:r-squared=" << r2 << endl;
  }
}

void tregclass::put_tregtab1(ostream& fout, double sp, int dfp, double sl, int dfl, double sa, int dfa, double sr, int dfr, double st, int dft) const
// 一元配置型分散分析表を出力する
{
  constexpr int LENline = 78;
  string tstr("ANOVA1 Table");
  fout << setw((LENline - strclen(tstr, CHARwide)) / 2 + tstr.size()) << tstr;
  fout << "\nFACTOR    Sum of Squares        DF  Mean Squares     F-value       p-value\n";
  putline(fout, LENline);
  double v = 0.0, f = 0.0, pr = 0.0;
  double vr = calvfp(v, f, pr, sp, dfp, sr, dfr);
  ios::fmtflags flags0 = fout.flags(); // 現在の書式フラッグを保存する
  tstr = "REGRESSION";
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sp;
  fout << setw(10) << dfp << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  calvfp(v, f, pr, sl, dfl, sr, dfr);
  tstr = "LOF";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sl;
  fout << setw(10) << dfl << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, sa, dfa, sr, dfr);
  tstr = "PERIOD";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sa;
  fout << setw(10) << dfa << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  tstr = "RESIDUALS";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sr;
  fout << setw(10) << dfr << setw(14) << vr;
  fout << "\n";
  putline(fout, LENline);
  tstr = "TOTAL";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << st;
  fout << setw(10) << dft << endl;
}

void tregclass::put_tregtab2(ostream& fout, double sp, int dfp, double sl, int dfl, double sa, int dfa, double sb, int dfb, double sr, int dfr, double st, int dft) const
// 二元配置型分散分析表を出力する
{
  constexpr int LENline = 78;
  string tstr("ANOVA2 Table");
  fout << setw((LENline - strclen(tstr, CHARwide)) / 2 + tstr.size()) << tstr;
  fout << "\nFACTOR    Sum of Squares        DF  Mean Squares     F-value       p-value\n";
  putline(fout, LENline);
  double v = 0.0, f = 0.0, pr = 0.0;
  double vr = calvfp(v, f, pr, sp, dfp, sr, dfr);
  tstr = "REGRESSION";
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sp;
  fout << setw(10) << dfp << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  calvfp(v, f, pr, sl, dfl, sr, dfr);
  tstr = "LOF";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sl;
  fout << setw(10) << dfl << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, sa, dfa, sr, dfr);
  tstr = "PERIOD";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sa;
  fout << setw(10) << dfa << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  calvfp(v, f, pr, sb, dfb, sr, dfr);
  tstr = "SUBJECT";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sb;
  fout << setw(10) << dfb << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  tstr = "RESIDUALS";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sr;
  fout << setw(10) << dfr << setw(14) << vr;
  fout << "\n";
  putline(fout, LENline);
  tstr = "TOTAL";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << st;
  fout << setw(10) << dft << endl;
}

void tregclass::plot_ngp(const varclass& var) // Ngraph用データを出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("mse"))           // 平均値±SE
  {
    fout << "-s1 -dmark,00 -o300\n";
    plot_ngpid(fout);
    plot_ngpmse(fout, var);
    delml = true;
  }
  if (opt.ngp("reg"))           // 回帰直線
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dline -w80\n";
    plot_ngpid(fout);
    plot_ngpreg(fout, var);
  }
}

void tregclass::plot_ngpid(ofstream& fout)
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);
  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%PERIOD ITEM  1:" << id.item(0) << "\n";
  fout << "%  :\n";
  fout << "%PERIOD ITEM" << setw(3) << id.ni();
  fout << ":" << id.item(id.ni() - 1) << "\n";
}

void tregclass::plot_ngpmse(ofstream& fout, const varclass& var) const
// Ngraph用時期別平均±SEを出力する
{
  fout << "%<MEAN +/- SE(when draw SE:-derrorbar_y -w40 -o500)>\n";
  for (size_t j = 0; j < id.ni(); ++j)
  {
    fout << opt.t(j) << " " << var.mean(j) << " " << -var.se(j) << " " << var.se(j) << "\n";
  }
}

void tregclass::plot_ngpreg(ofstream& fout, const varclass& var) const
// Ngraph時系列回帰直線を出力する
{
  fout << "%<TIIME SERIES REGRESSION LINE>\n";
  int p = id.ni();
  double dx = (opt.t(p - 1) - opt.t(0)) * 0.1;
  double x = opt.t(0) - dx;
  fout << x << " " << var.b0() + var.b1() * x << "\n";
  x = opt.t(p - 1) + dx;
  fout << x << " " << var.b0() + var.b1() * x << endl;
}

int tregclass::get(const char* c_fname) // データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void tregclass::put(int nc)     // 集計結果を出力する
{
  if (nc < 1)                   // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout);
      cout << "\n" << endl;
    }
    return;
  }
  varclass var(id.ni());        // 変数オブジェクト
  if (var.cal(opt, datalst) < 1) return; // 計算できず
  if (opt.res("res"))
  {
    put_resid(cout);
    if (opt.res("stat")) put_stat(cout, var);
    if (opt.res("reg"))
    {
      put_tregfun(cout, var);
      if (var.st() > 0.0)
      {
        cout << "\n";
        switch (opt.am())
        {
        case am0 : putatab(cout, "REGRESSION", var.sp(), var.dfp(), var.sr(), var.dfr(), var.st(), var.dft());
          break;
        case am1 : put_tregtab1(cout, var.sp(), var.dfp(), var.sl(), var.dfl(), var.sa(), var.dfa(), var.sr(), var.dfr(), var.st(), var.dft());
          break;
        default : put_tregtab2(cout, var.sp(), var.dfp(), var.sl(), var.dfl(), var.sa(), var.dfa(), var.sb(), var.dfb(), var.sr(), var.dfr(), var.st(), var.dft());
        }
      }
    }
    cout << "\n" << endl;
  }
  if (opt.ngpname() != "") plot_ngp(var);
}

int main(int argc, char* argv[]) // メインルメーチン
{
  tregclass treg(argv[0]);      // メインクラス変数
  try
  {
    treg.init();                // tregの初期設定
    bool fok = true;            // ファイル処理フラッグ
    int nc = 0;                 // 読み込んだケース数
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) nc += treg.get(nullptr); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { treg.setopt(argv[i] + 1);  fok = true; }
      else
      { nc += treg.get(argv[i]);  fok = false; }
    }
    treg.put(nc);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(treg.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(treg.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
